# SPDX-License-Identifier: MIT

import gc
import os
import tempfile
import unittest

from ..utility import (
    QtTest,
    click_button,
    delete_folder_if_exists,
)

import numpy as np

from qtpy import QtWidgets
from mock import MagicMock
from xypattern import Pattern

from ...model.DioptasModel import DioptasModel
from ...widgets.integration import IntegrationWidget
from ...controller.integration import IntegrationController

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, os.pardir, "data")
jcpds_path = os.path.join(data_path, "jcpds")


class IntegrationFunctionalTest(QtTest):
    def setUp(self):
        self.model = DioptasModel()

        self.integration_widget = IntegrationWidget()
        self.integration_controller = IntegrationController(
            widget=self.integration_widget, dioptas_model=self.model
        )
        self.model.calibration_model.load(
            os.path.join(data_path, "CeO2_Pilatus1M.poni")
        )
        self.model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))

        self.integration_pattern_controller = (
            self.integration_controller.pattern_controller
        )
        self.integration_image_controller = self.integration_controller.image_controller

    def test_activating_mask_mode(self):
        y1 = self.model.pattern.y

        self.model.mask_model.mask_below_threshold(self.model.img_data, 1)
        click_button(self.integration_widget.img_mask_btn)
        y2 = self.model.pattern.y
        self.assertFalse(np.array_equal(y1, y2))

        click_button(self.integration_widget.img_mask_btn)
        y3 = self.model.pattern.y
        self.assertTrue(np.array_equal(y1, y3))

    def test_activating_roi_mode(self):
        y1 = self.model.pattern.y

        click_button(self.integration_widget.img_roi_btn)
        self.assertIsNotNone(self.model.current_configuration.mask_model.roi_mask)

        y2 = self.model.pattern.y
        self.assertFalse(np.array_equal(y1, y2))

        click_button(self.integration_widget.img_roi_btn)
        y3 = self.model.pattern.y
        self.assertTrue(np.array_equal(y1, y3))

    def test_activating_roi_mode_and_mask_mode(self):
        y1 = self.model.pattern.y

        self.model.mask_model.mask_below_threshold(self.model.img_data, 1)
        click_button(self.integration_widget.img_mask_btn)
        y2 = self.model.pattern.y

        click_button(self.integration_widget.img_roi_btn)
        y3 = self.model.pattern.y

        click_button(self.integration_widget.img_roi_btn)
        y4 = self.model.pattern.y

        click_button(self.integration_widget.img_mask_btn)
        y5 = self.model.pattern.y

        self.assertFalse(np.array_equal(y3, y1))
        self.assertFalse(np.array_equal(y3, y2))
        self.assertFalse(np.array_equal(y3, y4))

        self.assertFalse(np.array_equal(y1, y2))
        self.assertFalse(np.array_equal(y1, y4))
        self.assertFalse(np.array_equal(y1, y3))
        self.assertTrue(np.array_equal(y1, y5))

    def test_moving_roi(self):
        click_button(self.integration_widget.img_roi_btn)
        roi_limits1 = self.integration_widget.img_widget.roi.getRoiLimits()
        y1 = self.model.pattern.y
        self.integration_widget.img_widget.roi.setX(30)
        self.integration_widget.img_widget.roi.setPos((31, 31))
        self.integration_widget.img_widget.roi.setSize((100, 100))
        roi_limits2 = self.integration_widget.img_widget.roi.getRoiLimits()
        y2 = self.model.pattern.y

        self.assertNotEqual(roi_limits1, roi_limits2)
        self.assertFalse(np.array_equal(y1, y2))

    def test_changing_integration_unit(self):
        x1, y1 = self.model.pattern.data

        click_button(self.integration_widget.pattern_q_btn)
        x2, y2 = self.model.pattern.data
        self.assertLess(np.max(x2), np.max(x1))

        click_button(self.integration_widget.pattern_d_btn)
        x3, y3 = self.model.pattern.data
        self.assertGreater(np.max(x3), np.max(x2))

        click_button(self.integration_widget.pattern_tth_btn)
        x4, y4 = self.model.pattern.data
        self.assertTrue(np.array_equal(x1, x4))
        self.assertTrue(np.array_equal(y1, y4))

    def test_configuration_selected_changes_img_mode(self):
        click_button(self.integration_widget.img_mode_btn)
        self.assertEqual(self.integration_widget.img_mode, "Cake")
        self.assertTrue(self.model.current_configuration.auto_integrate_cake)

        self.model.add_configuration()
        self.model.select_configuration(0)
        self.assertEqual(self.integration_widget.img_mode, "Cake")
        self.model.select_configuration(1)
        self.assertFalse(self.model.current_configuration.auto_integrate_cake)
        self.assertEqual(self.integration_widget.img_mode, "Image")

    def test_configuration_selected_changes_green_line_position_in_image_mode(self):
        self.integration_image_controller.img_mouse_click(0, 500)
        self.model.add_configuration()
        self.model.calibration_model.load(
            os.path.join(data_path, "CeO2_Pilatus1M_2.poni")
        )
        self.model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
        self.integration_image_controller.img_mouse_click(0, 500)
        self.model.select_configuration(0)

    def test_configuration_selected_changes_green_line_position_in_cake_mode(self):
        self.integration_image_controller.img_mouse_click(350, 100)
        click_button(self.integration_widget.img_mode_btn)
        self.model.add_configuration()
        self.model.calibration_model.load(
            os.path.join(data_path, "CeO2_Pilatus1M_2.poni")
        )
        click_button(self.integration_widget.img_mode_btn)
        self.model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
        self.integration_image_controller.img_mouse_click(1840, 500)
        self.model.select_configuration(0)

    def test_azimuthal_plot_shows_same_independent_of_unit(self):
        click_button(self.integration_widget.img_mode_btn)
        self.integration_image_controller.img_mouse_click(600, 150)
        x1, y1 = self.integration_widget.cake_widget.cake_integral_item.getData()
        click_button(self.integration_widget.pattern_q_btn)
        self.integration_image_controller.img_mouse_click(600, 150)
        x2, y2 = self.integration_widget.cake_widget.cake_integral_item.getData()
        self.assertAlmostEqual(np.nansum((x1 - x2) ** 2), 0)
        self.assertAlmostEqual(np.nansum((y1 - y2) ** 2), 0)


class BatchIntegrationFunctionalTest(QtTest):
    def setUp(self):
        self.model = DioptasModel()

        self.integration_widget = IntegrationWidget()
        self.integration_controller = IntegrationController(
            widget=self.integration_widget, dioptas_model=self.model
        )

        pattern = Pattern(
            x=np.linspace(0, 35, 2500), y=np.random.rand(2500) * 1e4
        )
        self.model.calibration_model.integrate_1d = MagicMock(
            return_value=(pattern.x, pattern.y)
        )

        self.model.calibration_model.is_calibrated = True

        self.model.calibration_model.load(os.path.join(data_path, "lambda/L2.poni"))

        files = [
            os.path.join(data_path, "lambda/testasapo1_1009_00002_m1_part00000.nxs"),
            os.path.join(data_path, "lambda/testasapo1_1009_00002_m1_part00001.nxs"),
        ]

        QtWidgets.QFileDialog.getOpenFileNames = MagicMock(return_value=files)
        click_button(self.integration_widget.batch_widget.file_control_widget.load_btn)

        self.integration_controller.batch_controller.integrate()

    def tearDown(self):
        del self.integration_widget
        del self.integration_controller
        del self.model
        gc.collect()

        # delete_folder_if_exists(os.path.join(data_path, 'lambda_temp'))

    def test_data_is_ok(self):
        self.assertTrue(self.model.batch_model.data is not None)
        self.assertEqual(self.model.batch_model.data.shape[0], 20)
        self.assertEqual(
            self.model.batch_model.data.shape[1],
            self.model.batch_model.binning.shape[0],
        )
        self.assertEqual(
            self.model.batch_model.data.shape[0], self.model.batch_model.n_img
        )
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 19)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(20/20):")

    def save_pattern(self, filename):
        QtWidgets.QFileDialog.getSaveFileName = MagicMock(return_value=filename)
        click_button(self.integration_widget.batch_widget.file_control_widget.save_btn)

    def test_save_data(self):
        for ext in ["nxs", "csv", "png"]:
            self.save_pattern(os.path.join(data_path, f"Test_spec.{ext}"))
            self.assertTrue(os.path.exists(os.path.join(data_path, f"Test_spec.{ext}")))
            self.assertGreater(
                os.stat(os.path.join(data_path, f"Test_spec.{ext}")).st_size, 1
            )
            os.remove(os.path.join(data_path, f"Test_spec.{ext}"))

        self.save_pattern(os.path.join(data_path, "Test_spec.dat"))
        for i in range(20):
            filepath = os.path.join(data_path, f"Test_spec_{i:03d}.dat")
            self.assertTrue(os.path.exists(filepath))
            self.assertGreater(os.stat(filepath).st_size, 1)
            os.remove(filepath)

    def test_save_load_reintegrate(self):
        tmp = tempfile.NamedTemporaryFile(suffix=".nxs", delete=False)
        tmp.close()
        self.addCleanup(os.unlink, tmp.name)
        self.save_pattern(tmp.name)

        QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
            return_value=[tmp.name]
        )
        click_button(self.integration_widget.batch_widget.file_control_widget.load_btn)

        self.assertEqual(self.model.batch_model.data.shape[0], 20)
        self.assertEqual(
            self.model.batch_model.data.shape[1],
            self.model.batch_model.binning.shape[0],
        )
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 19)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(20/20):")

        self.integration_widget.batch_widget.position_widget.step_series_widget.step_txt.setValue(
            2
        )
        self.integration_controller.batch_controller.integrate()
        self.assertEqual(self.model.batch_model.data.shape[0], 10)
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 9)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(10/20):")

    def test_integrate_with_parameters(self):

        self.integration_widget.batch_widget.position_widget.step_series_widget.step_txt.setValue(
            1
        )
        self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.setValue(
            4
        )
        self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.setValue(
            15
        )
        self.integration_controller.batch_controller.integrate()
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(self.model.batch_model.data.shape[0], 12)
        self.assertEqual(stop, 11)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(12/20):")

        self.integration_widget.batch_widget.position_widget.step_series_widget.step_txt.setValue(
            2
        )
        self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.setValue(
            0
        )
        self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.setValue(
            11
        )
        self.integration_controller.batch_controller.integrate()
        self.assertEqual(self.model.batch_model.data.shape[0], 6)
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 5)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(6/20):")

    @unittest.skip("Test is not useful - feature will be disabled in future")
    def test_load_missing_raw(self):
        """
        Load processed data, while raw data not available.

        Fix error by giving correct path to raw data
        """

        # Create tmp raw data
        import shutil

        shutil.copytree(
            os.path.join(data_path, "lambda"), os.path.join(data_path, "lambda_temp")
        )
        # Integrate tmp data. Save proc. Delete tmp data.
        files = [
            os.path.join(
                data_path, "lambda_temp/testasapo1_1009_00002_m1_part00000.nxs"
            ),
            os.path.join(
                data_path, "lambda_temp/testasapo1_1009_00002_m1_part00001.nxs"
            ),
        ]

        # QtWidgets.QFileDialog.getOpenFileNames = MagicMock(return_value=files)
        # click_button(self.integration_widget.batch_widget.file_control_widget.load_btn)

        self.integration_controller.batch_controller.integrate()
        self.save_pattern(os.path.join(data_path, f"Test_missing_raw.nxs"))

        delete_folder_if_exists(os.path.join(data_path, "lambda_temp"))
        # Load proc data with missing raw data
        QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
            return_value=[os.path.join(data_path, "Test_missing_raw.nxs")]
        )
        click_button(self.integration_widget.batch_widget.file_control_widget.load_btn)
        self.assertTrue(self.model.batch_model.raw_available is False)
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 19)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(20/20):")

        # Pattern widget is still working
        self.integration_controller.batch_controller.img_mouse_click(5, 15)
        x1, y1 = self.model.pattern.data
        y = self.model.batch_model.data[15]
        x = self.model.batch_model.binning
        self.assertTrue(np.array_equal(y1, y))
        self.assertTrue(np.array_equal(x1, x))

        # Fix raw data path
        self.model.working_directories["image"] = os.path.join(data_path, "lambda")
        QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
            return_value=[os.path.join(data_path, "Test_missing_raw.nxs")]
        )
        click_button(self.integration_widget.batch_widget.load_btn)
        self.assertEqual(self.model.batch_model.n_img_all, 20)
        self.assertTrue(self.model.batch_model.raw_available)
        self.assertEqual(
            os.path.basename(self.model.batch_model.calibration_model.filename),
            "L2.poni",
        )
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 19)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(20/20):")

        # Cleanup
        os.remove(os.path.join(data_path, f"Test_missing_raw.nxs"))

    def test_create_waterfall(self):
        click_button(self.integration_widget.batch_widget.mode_widget.view_2d_btn)
        click_button(self.integration_widget.batch_widget.control_widget.waterfall_btn)
        QtWidgets.QApplication.processEvents()

        # Create waterfall
        self.integration_controller.batch_controller.img_mouse_click(5, 5)
        self.integration_controller.batch_controller.rect.set_size(10, 15)
        self.integration_controller.batch_controller.img_mouse_click(10, 15)
        self.assertEqual(len(self.model.overlay_model.overlays), 10)

        # edit waterfall
        self.integration_widget.batch_widget.position_widget.step_series_widget.step_txt.setValue(
            2
        )
        self.integration_controller.batch_controller.process_step()
        self.assertEqual(len(self.model.overlay_model.overlays), 5)

    def test_show_phases(self):
        click_button(self.integration_widget.batch_widget.mode_widget.view_2d_btn)

        self.model.phase_model.add_jcpds(os.path.join(jcpds_path, "FeGeO3_cpx.jcpds"))
        click_button(self.integration_widget.batch_widget.control_widget.phases_btn)

        self.assertEqual(
            len(self.integration_widget.batch_widget.stack_plot_widget.img_view.phases),
            1,
        )
        self.assertEqual(
            len(
                self.integration_widget.batch_widget.stack_plot_widget.img_view.phases[
                    0
                ].line_items
            ),
            27,
        )

        last_line_position = (
            self.integration_widget.batch_widget.stack_plot_widget.img_view.phases[0]
            .line_items[-1]
            .getPos()
        )
        self.assertGreater(last_line_position[0], 900)

    def test_change_view(self):
        self.integration_widget.batch_widget.position_widget.step_series_widget.step_txt.setValue(
            2
        )
        self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.setValue(
            0
        )
        self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.setValue(
            12
        )
        self.integration_controller.batch_controller.integrate()
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 6)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(7/20):")

        self.integration_widget.batch_widget.mode_widget.view_f_btn.setChecked(True)
        self.integration_controller.batch_controller.change_view()
        start = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.integration_widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        frame = str(
            self.integration_widget.batch_widget.position_widget.step_series_widget.pos_label.text()
        )
        self.assertEqual(stop, 6)
        self.assertEqual(start, 0)
        self.assertEqual(frame, "Frame(7/20):")

    def test_change_unit(self):
        self.integration_controller.batch_controller.img_mouse_click(5, 15)
        x1, y1 = self.model.pattern.data
        y = self.model.batch_model.data[15]
        x = self.model.batch_model.binning
        self.assertTrue(np.array_equal(y1, y))
        self.assertTrue(np.array_equal(x1, x))
        self.assertGreater(x1[-1], x1[0])

        click_button(self.integration_widget.batch_widget.options_widget.q_btn)
        self.model.calibration_model.integrate_1d.assert_called_with(
            mask=None, num_points=None, unit="q_A^-1", azi_range=None, trim_zeros=True
        )

        click_button(self.integration_widget.batch_widget.options_widget.d_btn)
        self.model.calibration_model.integrate_1d.assert_called_with(
            mask=None, num_points=None, unit="d_A", azi_range=None, trim_zeros=True
        )

        click_button(self.integration_widget.batch_widget.options_widget.tth_btn)
        self.model.calibration_model.integrate_1d.assert_called_with(
            mask=None, num_points=None, unit="2th_deg", azi_range=None, trim_zeros=True
        )



# SPDX-License-Identifier: MIT

import os

import numpy as np
import pytest
from mock import MagicMock
from xypattern import Pattern
from xypattern.auto_background import SmoothBrucknerBackground

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, "../data")


def test_add_configuration(dioptas_model):
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))

    prev_sum = np.sum(dioptas_model.img_data)
    assert np.array_equal(
        dioptas_model.img_data, dioptas_model.configurations[0].img_model.img_data
    )

    dioptas_model.add_configuration()
    new_sum = np.sum(dioptas_model.img_data)
    assert np.array_equal(
        dioptas_model.img_data, dioptas_model.configurations[1].img_model.img_data
    )

    assert prev_sum == new_sum


def test_remove_configuration(dioptas_model):
    dioptas_model.add_configuration()
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))
    old_img = dioptas_model.img_data

    dioptas_model.remove_configuration()
    assert not np.array_equal(dioptas_model.img_data, old_img)


def test_select_configuration(dioptas_model):
    img_1 = dioptas_model.img_data

    dioptas_model.add_configuration()
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))
    img_2 = dioptas_model.img_data

    dioptas_model.add_configuration()
    dioptas_model.img_model.load(os.path.join(data_path, "image_002.tif"))
    img_3 = dioptas_model.img_data

    dioptas_model.select_configuration(0)
    assert np.array_equal(dioptas_model.img_data, img_1)

    dioptas_model.select_configuration(2)
    assert np.array_equal(dioptas_model.img_data, img_3)

    dioptas_model.select_configuration(1)
    assert np.array_equal(dioptas_model.img_data, img_2)


def test_signals_are_raised(dioptas_model):
    dioptas_model.configuration_added = MagicMock()
    dioptas_model.configuration_selected = MagicMock()
    dioptas_model.configuration_removed = MagicMock()

    dioptas_model.add_configuration()
    dioptas_model.add_configuration()
    dioptas_model.configuration_added.emit.assert_called()

    dioptas_model.select_configuration(0)
    dioptas_model.configuration_selected.emit.assert_called_with(0)

    dioptas_model.remove_configuration()
    dioptas_model.configuration_removed.emit.assert_called_with(0)


def test_integrate_cakes(dioptas_model):
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.current_configuration.auto_integrate_cake = True
    dioptas_model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    assert not np.array_equal(
        dioptas_model.current_configuration.cake_img, np.zeros((2048, 2048))
    )


def test_integrate_cake_with_mask(dioptas_model):
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.current_configuration.auto_integrate_cake = True
    dioptas_model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    cake_img1 = dioptas_model.current_configuration.cake_img

    dioptas_model.use_mask = True
    dioptas_model.mask_model.mask_below_threshold(dioptas_model.img_data, 1)
    dioptas_model.img_model.img_changed.emit()
    cake_img2 = dioptas_model.current_configuration.cake_img
    assert not np.array_equal(cake_img1, cake_img2)


def test_integrate_cake_with_different_azimuth_points(dioptas_model):
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.current_configuration.auto_integrate_cake = True
    dioptas_model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))

    assert dioptas_model.current_configuration.cake_img.shape[0] == 360
    dioptas_model.current_configuration.cake_azimuth_points = 720
    assert dioptas_model.current_configuration.cake_img.shape[0] == 720


def test_integrate_cake_with_different_rad_points(dioptas_model):
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.current_configuration.auto_integrate_cake = True
    dioptas_model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))

    assert dioptas_model.current_configuration.cake_img.shape[1] > 360
    dioptas_model.current_configuration.integration_rad_points = 720
    assert dioptas_model.current_configuration.cake_img.shape[1], 720


def test_change_cake_azimuth_range(dioptas_model):
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.current_configuration.auto_integrate_cake = True
    dioptas_model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))

    dioptas_model.current_configuration.cake_azimuth_range = [-180, 180]

    assert dioptas_model.current_configuration.calibration_model.cake_azi[
        0
    ] == pytest.approx(-179.5)
    assert dioptas_model.current_configuration.calibration_model.cake_azi[
        -1
    ] == pytest.approx(179.5)

    dioptas_model.current_configuration.cake_azimuth_range = [-100, 100]
    assert dioptas_model.current_configuration.calibration_model.cake_azi[0] > -100
    assert dioptas_model.current_configuration.calibration_model.cake_azi[-1] < 100


def prepare_combined_patterns(model):
    x1 = np.linspace(0, 10)
    y1 = np.ones(x1.shape)
    pattern1 = Pattern(x1, y1)

    x2 = np.linspace(7, 15)
    y2 = np.ones(x2.shape) * 2
    pattern2 = Pattern(x2, y2)

    model.pattern_model.pattern = pattern1
    model.add_configuration()
    model.pattern_model.pattern = pattern2

    model.combine_patterns = True


def test_combine_patterns(dioptas_model):
    prepare_combined_patterns(dioptas_model)

    assert dioptas_model.pattern is not None
    x3, y3 = dioptas_model.pattern.data
    assert np.min(x3) < 7
    assert np.max(x3) > 10


def test_save_combine_patterns(dioptas_model, tmp_path):
    prepare_combined_patterns(dioptas_model)
    file_path = os.path.join(tmp_path, "combined_pattern.xy")
    dioptas_model.pattern.save(file_path)
    saved_pattern = Pattern.from_file(file_path)
    x3, y3 = saved_pattern.data
    assert np.min(x3) < 7
    assert np.max(x3) > 10


def test_combine_cakes(dioptas_model):
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.current_configuration.auto_integrate_cake = True
    dioptas_model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))

    cake1 = dioptas_model.cake_data
    dioptas_model.add_configuration()

    dioptas_model.calibration_model.load(
        os.path.join(data_path, "CeO2_Pilatus1M_2.poni")
    )
    dioptas_model.current_configuration.auto_integrate_cake = True
    dioptas_model.img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))

    cake2 = dioptas_model.cake_data
    dioptas_model.combine_cakes = True
    assert not np.array_equal(dioptas_model.cake_data, cake1)
    assert not np.array_equal(dioptas_model.cake_data, cake2)


def test_setting_factors(dioptas_model):
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))
    data1 = np.copy(dioptas_model.img_data)
    dioptas_model.img_model.factor = 2
    assert np.array_equal(2 * data1, dioptas_model.img_data)


def test_iterate_next_image(dioptas_model):
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))
    dioptas_model.add_configuration()
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))

    dioptas_model.next_image()

    assert dioptas_model.configurations[0].img_model.filename == os.path.abspath(
        os.path.join(data_path, "image_002.tif")
    )
    assert dioptas_model.configurations[1].img_model.filename == os.path.abspath(
        os.path.join(data_path, "image_002.tif")
    )


def test_iterate_previous_image(dioptas_model):
    dioptas_model.img_model.load(os.path.join(data_path, "image_002.tif"))
    dioptas_model.add_configuration()
    dioptas_model.img_model.load(os.path.join(data_path, "image_002.tif"))

    dioptas_model.previous_image()

    assert dioptas_model.configurations[0].img_model.filename == os.path.abspath(
        os.path.join(data_path, "image_001.tif")
    )
    assert dioptas_model.configurations[1].img_model.filename == os.path.abspath(
        os.path.join(data_path, "image_001.tif")
    )


def test_unit_change_with_auto_background_subtraction(dioptas_model):
    # load calibration and image
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))

    # check that background subtraction works
    x, y = dioptas_model.pattern_model.pattern.data
    x_max_2th = np.max(x)
    roi = (0, np.max(x) + 1)
    dioptas_model.pattern_model.set_auto_background_subtraction((0.1, 50, 50), roi)
    new_y = dioptas_model.pattern_model.pattern.y
    assert np.sum(y - new_y) != 0

    x_bkg, _ = dioptas_model.pattern_model.pattern.auto_background_pattern.data

    # change the unit to q
    dioptas_model.integration_unit = "q_A^-1"

    # check that the pattern is integrated with different unit
    x, y = dioptas_model.pattern_model.pattern.data
    x_max_q = np.max(x)
    assert x_max_q < x_max_2th

    auto_bkg = dioptas_model.pattern_model.pattern.auto_bkg
    assert type(auto_bkg) == SmoothBrucknerBackground
    assert auto_bkg.smooth_width < 0.1

    # check that the background roi has changed
    assert dioptas_model.pattern_model.pattern.auto_bkg_roi[1] != roi[1]
    assert dioptas_model.pattern_model.pattern.auto_bkg is not None

    # check that the background pattern has changed:
    x_bkg_2, _ = dioptas_model.pattern_model.pattern.auto_background_pattern.data
    assert np.max(x_bkg) != np.max(x_bkg_2)


def test_save_empty_configuration(dioptas_model, tmp_path):
    dioptas_model.save(os.path.join(tmp_path, "empty.dio"))


def test_clear_model(dioptas_model):
    dioptas_model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))

    dioptas_model.add_configuration()
    dioptas_model.add_configuration()

    dioptas_model.reset()



# SPDX-License-Identifier: MIT

import os

resources_path = os.path.join(os.path.dirname(__file__), 'resources')
calibrants_path = os.path.join(resources_path, 'calibrants')
icons_path = os.path.join(resources_path, 'icons')
data_path = os.path.join(resources_path, 'data')
style_path = os.path.join(resources_path, 'style')


# SPDX-License-Identifier: MIT
import gc

from dioptas.model.util.signal import Signal, WeakRefList


def test_weakreflist_with_objects():
    weak_ref_list = WeakRefList()

    class A(str):
        pass

    a = A()
    weak_ref_list.append(a)
    assert len(weak_ref_list) == 1
    assert weak_ref_list[0]() == a

    del a
    gc.collect()
    assert len(weak_ref_list) == 0


def test_weakreflist_with_bound_methods():
    weak_ref_list = WeakRefList()

    class A:
        def method(self):
            return "lala"

    a = A()
    weak_ref_list.append(a.method)

    assert len(weak_ref_list) == 1
    assert weak_ref_list[0]()() == "lala"

    del a
    gc.collect()
    assert len(weak_ref_list) == 0


def test_weakreflist_with_insertion():
    weak_ref_list = WeakRefList()

    class A:
        def method(self):
            return "lala"

    a = A()
    weak_ref_list.insert(0, a.method)

    assert len(weak_ref_list) == 1
    assert weak_ref_list[0]()() == "lala"

    del a
    gc.collect()
    assert len(weak_ref_list) == 0


def test_signal_with_sole_functions():
    signal = Signal()
    memory = []

    def f():
        memory.append(1)

    signal.connect(f)
    signal.emit()
    assert memory == [1]

    signal.disconnect(f)
    signal.emit()
    assert memory == [1]


def test_signal_with_bound_methods():
    signal = Signal()
    memory = []

    class A:
        def method(self):
            memory.append(1)

    a = A()
    signal.connect(a.method)
    signal.emit()
    assert memory == [1]

    signal.disconnect(a.method)
    signal.emit()
    assert memory == [1]


def test_signal_with_signal():
    signal = Signal()
    memory = []

    signal2 = Signal()

    def f():
        memory.append(1)

    signal.connect(f)
    signal2.connect(signal)
    signal2.emit()

    assert memory == [1]


def test_signal_with_priority():
    signal = Signal()
    memory = []

    def f():
        memory.append(1)

    def g():
        memory.append(2)

    signal.connect(f)
    signal.connect(g, priority=True)
    signal.emit()

    assert memory == [2, 1]



# SPDX-License-Identifier: MIT
from io import StringIO
import sys
import traceback

from qtpy import QtWidgets


def excepthook(exc_type, exc_value, traceback_obj):
    """
    Global function to catch unhandled exceptions. This function will result in an error dialog which displays the
    error information.

    :param exc_type: exception type
    :param exc_value: exception value
    :param traceback_obj: traceback object
    :return:
    """

    separator = '-' * 80

    tb_info_file = StringIO()
    traceback.print_tb(traceback_obj, None, tb_info_file)
    tb_info_file.seek(0)
    tb_info = tb_info_file.read()
    errmsg = '%s: \n%s' % (str(exc_type), str(exc_value))
    sections = [separator, separator, errmsg, separator, tb_info]
    msg = '\n'.join(sections)
    print(msg)
    # raise Exception


def main():
    app = QtWidgets.QApplication([])
    sys.excepthook = excepthook
    from sys import platform as _platform
    from ..controller.MainController import MainController

    if _platform == "linux" or _platform == "linux2" or _platform == "win32" or _platform == 'cygwin':
        app.setStyle('plastique')

    controller = MainController()
    controller.show_window()
    app.exec_()
    del app



# SPDX-License-Identifier: MIT
from __future__ import annotations

import os.path
import time

import h5py
import numpy as np
from dioptas.model.util.signal import Signal

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .Configuration import Configuration


class MapPointInfo:
    filename: str
    frame_index: int

    def __init__(self, filepath, frame_index=0):
        self.filepath = filepath
        self.filename = os.path.basename(filepath)
        self.frame_index = frame_index


class MapModel:
    map_changed = Signal()

    # point_integrated is emitted with the index of the integrated point
    # it will be a fractional number, when the image file contains multiple frames
    point_integrated = Signal(float)

    def __init__(self, configuration: "Configuration"):
        """
        Creates a new map-model. The configuration specified will serve as
        integrator for the processed files.
        :param configuration: Configuration to be used for integration
        """
        super().__init__()
        self.configuration = configuration
        self.filepaths = None
        self.point_infos = []
        self.pattern_intensities = None
        self.pattern_x = None
        self.window = None
        self.window_intensities = None
        self.dimension = None
        self.possible_dimensions = None
        self.map = None

    def load(self, filepaths: list[str], callback_fn=None):
        """Loads a list of files, integrates them and creates a map"""
        if len(filepaths) == 0:
            raise ValueError("No files to load")

        self.filepaths = filepaths

        self.integrate(callback_fn=callback_fn)

        if len(self.pattern_intensities) == 0:
            return

        if self.window is None:
            self.window = get_center_window(self.pattern_x)

        self.window_intensities = get_window_intensities(
            self.pattern_x, self.pattern_intensities, self.window
        )

        self.possible_dimensions = find_possible_dimensions(
            len(self.window_intensities)
        )

        if self.dimension is None or self.dimension not in self.possible_dimensions:
            self.dimension = self.possible_dimensions[0]

        self.map = create_map(self.window_intensities, self.dimension)
        self.map_changed.emit()

    def integrate(self, callback_fn=None):
        """Integrates all files in the filepaths list and stores the results"""
        if not self.configuration.calibration_model.is_calibrated:
            raise ValueError("Detector geometry is not calibrated")

        # initialize data structures
        self.pattern_x = []
        self.pattern_intensities = []
        self.point_infos = []

        # disable trimming trailing zeros for integration, otherwise the
        # integration will result in patterns with different length, which
        # will cause problems when creating the map
        trim_trailing_zeros_backup = self.configuration.trim_trailing_zeros
        self.configuration.trim_trailing_zeros = False

        self.configuration.img_model.img_changed.blocked = True

        try:
            self._integrate(callback_fn=callback_fn)
        except Exception as e:
            self._reset()
            raise e
        finally:
            # reset model to previous state
            self.configuration.trim_trailing_zeros = trim_trailing_zeros_backup
            self.configuration.img_model.img_changed.blocked = False

    def _integrate(self, callback_fn=None):
        cal = self.configuration.calibration_model
        unit = self.configuration.integration_unit
        azi_range = self.configuration.oned_azimuth_range

        if cal.can_use_dioptrin_batch(unit, azi_range):
            self._integrate_dioptrin_batch(callback_fn=callback_fn)
        else:
            self._integrate_pyFAI(callback_fn=callback_fn)

    def _integrate_pyFAI(self, callback_fn=None):
        frame_count = 0
        n_total = None
        last_callback_time = time.monotonic()

        for file_ind, filepath in enumerate(self.filepaths):
            self.configuration.img_model.load(filepath)

            series_max = self.configuration.img_model.series_max
            if n_total is None:
                n_total = len(self.filepaths) * series_max

            for frame_ind in range(series_max):
                self.configuration.img_model.load_series_img(frame_ind + 1)
                x, y = self.configuration.integrate_image_1d(
                    update_pattern_model=False
                )

                if file_ind == 0 and frame_ind == 0:
                    self.pattern_x = x
                else:
                    if len(x) != len(self.pattern_x):
                        raise ValueError(
                            "The integrated patterns have different length, this is not supported"
                        )

                self.point_infos.append(MapPointInfo(filepath, frame_ind))
                self.pattern_intensities.append(y)

                frame_count += 1
                self.point_integrated.emit(
                    file_ind + (frame_ind + 1) / series_max
                )

                now = time.monotonic()
                if callback_fn is not None and now - last_callback_time > 0.1:
                    last_callback_time = now
                    if not callback_fn(frame_count, n_total):
                        self.pattern_intensities = np.array(self.pattern_intensities)
                        return

        # Final callback to ensure progress reaches 100%
        if callback_fn is not None:
            callback_fn(frame_count, n_total)

        self.pattern_intensities = np.array(self.pattern_intensities)

    def _integrate_dioptrin_batch(self, callback_fn=None):
        """Load frames through ImgModel, integrate via batch1d_iter with generator."""
        from dioptas.model.util.integration import iter_frames_sequential, convert_tth_to_d

        cal = self.configuration.calibration_model
        unit = self.configuration.integration_unit
        num_points = self.configuration.integration_rad_points

        # Load first file to get shape and configure integrator.
        # The mask must be fetched AFTER loading so that
        # update_mask_dimension has resized it to match the image.
        self.configuration.img_model.load(self.filepaths[0])
        img_shape = self.configuration.img_model.img_data.shape
        first_series_max = self.configuration.img_model.series_max

        if self.configuration.use_mask:
            mask = self.configuration.mask_model.get_mask()
        elif self.configuration.mask_model.roi is not None:
            mask = self.configuration.mask_model.roi_mask
        else:
            mask = None

        num_points = cal.sync_dioptrin_for_batch(mask, unit, num_points, img_shape)

        # Estimate total frames from first file; refined if files differ
        n_total = len(self.filepaths) * first_series_max

        aborted = False
        # Built lazily by frame generator; consumed in result loop
        all_infos = []

        def frame_generator():
            yield from iter_frames_sequential(
                self.configuration.img_model,
                self.filepaths,
                img_shape=img_shape,
                abort_check=lambda: aborted,
                on_frame=lambda fp, fi: all_infos.append(MapPointInfo(fp, fi)),
            )

        result_iter = cal.dioptrin_batch1d_iter(frame_generator(), num_points)

        last_callback_time = time.monotonic()
        frame_count = 0
        for i, result in enumerate(result_iter):
            if aborted:
                # Drain remaining pre-fetched results from dioptrin without
                # processing them. This ensures the iterator is fully
                # consumed so cleanup doesn't crash on worker threads.
                continue

            if not result.is_ok():
                raise RuntimeError(
                    f"Dioptrin batch integration failed: {result.error}"
                )

            x = np.array(result.result.radial)
            y = np.array(result.result.intensity)
            frame_count = i + 1

            if frame_count == 1:
                self.pattern_x = x
            else:
                if len(x) != len(self.pattern_x):
                    raise ValueError(
                        "The integrated patterns have different length, "
                        "this is not supported"
                    )

            self.point_infos.append(all_infos[i])
            self.pattern_intensities.append(y)

            self.point_integrated.emit(frame_count)

            # Throttle callback to avoid GUI overhead dominating throughput
            now = time.monotonic()
            if callback_fn is not None and now - last_callback_time > 0.1:
                last_callback_time = now
                if not callback_fn(frame_count, n_total):
                    aborted = True
                    # Don't break — let the for loop drain remaining
                    # pre-fetched results so dioptrin shuts down cleanly.

        # Final callback to ensure progress reaches 100%
        if callback_fn is not None and not aborted:
            callback_fn(frame_count, n_total)

        self.pattern_intensities = np.array(self.pattern_intensities)

        if unit == "d_A":
            self.pattern_x = convert_tth_to_d(
                self.pattern_x, cal.pattern_geometry.wavelength
            )

    def set_integration_results(
        self,
        pattern_x,
        pattern_intensities,
        point_infos: list[MapPointInfo],
        filepaths: list[str],
    ):
        """Sets pre-computed integration results and rebuilds the map.

        This allows populating the model without re-integrating, e.g. when
        results were computed in a worker thread.
        """
        self.pattern_x = pattern_x
        self.pattern_intensities = pattern_intensities
        self.point_infos = point_infos
        self.filepaths = filepaths

        if self.window is None:
            self.window = get_center_window(self.pattern_x)

        self.window_intensities = get_window_intensities(
            self.pattern_x, self.pattern_intensities, self.window
        )

        self.possible_dimensions = find_possible_dimensions(
            len(self.window_intensities)
        )

        if self.dimension is None or self.dimension not in self.possible_dimensions:
            self.dimension = self.possible_dimensions[0]

        self.map = create_map(self.window_intensities, self.dimension)
        self.map_changed.emit()

    def _reset(self):
        self.filepaths = None
        self.point_infos = []
        self.pattern_intensities = None
        self.pattern_x = None
        self.dimension = None
        self.possible_dimensions = None
        self.map = None
        self.map_changed.emit()

    def set_window(self, window: tuple[float, float]):
        """Sets the window in the pattern for generating the map
        :param window: tuple/list of lower value and upper value of the window
        """
        self.window = window
        if self.pattern_x is None:
            return
        self.window_intensities = get_window_intensities(
            self.pattern_x, self.pattern_intensities, self.window
        )
        self.map = create_map(self.window_intensities, self.dimension)
        self.map_changed.emit()

    def set_dimension(self, dimension: tuple[float, float]):
        """Sets the dimension of the map"""
        if dimension not in self.possible_dimensions:
            return
        self.dimension = dimension
        self.map = create_map(self.window_intensities, self.dimension)
        self.map_changed.emit()

    def get_point_info(self, row_index: int, column_index: int) -> MapPointInfo | None:
        """Returns the point info for the specified row and column index"""
        if self.dimension is None:
            return None
        ind = self.get_point_index(row_index, column_index)
        if ind is None:
            return None
        if ind > len(self.point_infos) - 1:
            return None
        return self.point_infos[ind]

    def get_point_index(self, row_index: int, column_index: int) -> int | None:
        """Returns the point index inside the list of integrated images for the specified row and column index"""
        if self.dimension is None:
            return None
        return int(column_index + self.dimension[1] * row_index)

    def get_point_coordinates(self, index: int) -> tuple[int, int]:
        """Returns the row and column index for the specified point index"""
        if self.dimension is None:
            return None
        return divmod(index, self.dimension[1])

    def get_filenames(self) -> list[str]:
        """Returns a list of filenames for the integrated images, it will add the frame index if it is not 0"""
        filenames = []
        for point_info in self.point_infos:
            if point_info.frame_index == 0:
                filenames.append(point_info.filename)
            else:
                filenames.append(f"{point_info.filename}:{point_info.frame_index}")
        return filenames

    def select_point(self, row_index: int, column_index: int):
        """Selects the point at the specified row and column index, will trigger a load of the image through the
        configuration. Thus the image_changed signal will be sent to all listeners"""
        point_ind = self.get_point_index(row_index, column_index)
        if point_ind is None:
            return
        self.select_point_by_index(point_ind)

    def select_point_by_index(self, index: int):
        """Selects the point at the specified index (considering the list of images), will trigger a load of the
        image through the configuration. Thus the image_changed signal will be sent to all listeners
        """
        if index < 0 or index >= len(self.point_infos):
            return
        point_info = self.point_infos[index]
        self.configuration.img_model.load(
            point_info.filepath,
            point_info.frame_index,
        )

    def save_in_hdf5(self, hdf5_group):
        """Save map state into the given HDF5 group. Skips if no map data."""
        if self.filepaths is None:
            return

        g = hdf5_group.create_group("map")
        g.create_dataset("pattern_x", data=self.pattern_x)
        g.create_dataset("pattern_intensities", data=self.pattern_intensities)
        g.create_dataset("window", data=np.array(self.window, dtype="f8"))
        g.create_dataset("dimension", data=np.array(self.dimension, dtype="i"))

        dt = h5py.string_dtype()
        g.create_dataset("filepaths", data=self.filepaths, dtype=dt)

        pi_group = g.create_group("point_infos")
        pi_filepaths = [p.filepath for p in self.point_infos]
        pi_frame_indices = [p.frame_index for p in self.point_infos]
        pi_group.create_dataset("filepaths", data=pi_filepaths, dtype=dt)
        pi_group.create_dataset("frame_indices", data=pi_frame_indices)

    def load_from_hdf5(self, hdf5_group):
        """Restore map state from the given HDF5 group."""
        if "map" not in hdf5_group:
            return

        g = hdf5_group["map"]
        pattern_x = g["pattern_x"][...]
        pattern_intensities = g["pattern_intensities"][...]
        filepaths = [fp.decode() if isinstance(fp, bytes) else fp for fp in g["filepaths"][...]]

        pi_fps = g["point_infos"]["filepaths"][...]
        pi_fis = g["point_infos"]["frame_indices"][...]
        point_infos = [
            MapPointInfo(
                fp.decode() if isinstance(fp, bytes) else fp,
                int(fi),
            )
            for fp, fi in zip(pi_fps, pi_fis)
        ]

        window = tuple(g["window"][...])
        dimension = tuple(g["dimension"][...])

        # Set window and dimension before set_integration_results so it
        # uses the saved values instead of computing defaults that may not
        # match the data after an HDF5 round-trip.
        self.window = window
        self.dimension = dimension

        # Block map_changed during restore — it is a class-level signal
        # shared by all instances, so emitting here would trigger handlers
        # connected via other instances (e.g. MapController) while
        # DioptasModel.configurations is still empty during load.
        self.map_changed.blocked = True
        try:
            self.set_integration_results(
                pattern_x, pattern_intensities, point_infos, filepaths
            )
        finally:
            self.map_changed.blocked = False


def get_center_window(x, window_range=3) -> list[float]:
    """
    Estimates a window of [x_min, x_max] centered in the x value list.
    :param x: a numpy array
    :param window_range: the window will be estimated with +- range * x_step
    :return: windows with [x_min, x_max]
    """
    window_center = x[int(len(x) / 2)]
    x_step = np.mean(np.diff(x))
    return [
        window_center - window_range * x_step,
        window_center + window_range * x_step,
    ]


def ind_in_window(x_array, window: tuple[float, float]) -> np.ndarray:
    """
    Gets the indices of a numpy array which are in the window
    :param x_array: a numpy array
    :param window: tuple/list of lower value and upper value of the window
    :return: list of indices
    """
    return np.where((x_array > window[0]) & (x_array < window[1]))[0]


def get_window_intensities(
    pattern_x, intensities, window: tuple[float, float]
) -> np.ndarray:
    """
    Estimates the intensities inside the specified window
    :param pattern_x: a numpy array of x values from the pattern
    :param intensities: a 2D numpy array holding the intensities of all patterns
    :param window: tuple/list of lower value and upper value of the summing window
    :return: an 1D array containing the sum of  intensities inside the window for each pattern
    """
    indices = ind_in_window(pattern_x, window)
    return np.sum(intensities[:, indices], axis=1)


def find_possible_dimensions(num_points: int) -> list[tuple[int, int]]:
    """
    Finds the possible dimension for a map with a given number of points
    :param num_points: number of points for the map
    :return: list of dimension pairs (x-dimension, y-dimension)
    """
    dimension_pairs = []
    for n in range(1, int(np.floor(np.sqrt(num_points + 1))) + 1):
        if num_points % n == 0:
            dim1 = n
            dim2 = num_points // n
            dimension_pairs.append((dim1, dim2))
            if dim1 != dim2:
                dimension_pairs.append((dim2, dim1))
    dimension_pairs.sort(key=lambda x: ((x[0] + x[1]) / 2 - np.sqrt(num_points)) ** 2)
    return dimension_pairs


def create_map(data: np.ndarray, dimension: tuple[int, int]) -> np.ndarray:
    """
    Creates a new map from the given 1D array and specified dimension. It will
    always create a copy of the data.
    :param data: input data for creating the map
    :param dimension: tuple of integers giving the shape of the map
    :return: numpy array with the reorganized data
    """
    new_data = np.copy(data)
    return np.reshape(new_data, dimension)





# SPDX-License-Identifier: MIT

from ..utility import QtTest, click_button, click_checkbox
import os
import gc

from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtTest import QTest
from mock import MagicMock

from ...controller.integration import PatternController
from ...controller.integration import PhaseController
from ...model.DioptasModel import DioptasModel
from ...widgets.integration import IntegrationWidget

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, "../data")
jcpds_path = os.path.join(data_path, "jcpds")


class PhaseControllerTest(QtTest):
    def setUp(self):
        self.model = DioptasModel()
        self.model.calibration_model.is_calibrated = True
        self.model.calibration_model.pattern_geometry.wavelength = 0.31e-10
        self.model.calibration_model.integrate_1d = MagicMock(
            return_value=(
                self.model.calibration_model.tth,
                self.model.calibration_model.int,
            )
        )
        self.widget = IntegrationWidget()
        self.widget.pattern_widget._auto_range = True
        self.phase_tw = self.widget.phase_tw
        self.phase_widget = self.widget.phase_widget

        self.pattern_controller = PatternController(self.widget, self.model)
        self.controller = PhaseController(self.widget, self.model)
        self.model.pattern_model.load_pattern(os.path.join(data_path, "pattern_001.xy"))

    def tearDown(self):
        del self.pattern_controller
        del self.controller
        del self.widget
        self.model.delete_configurations()
        del self.model
        gc.collect()

    def test_manual_deleting_phases(self):
        self.load_phases()

        self.assertEqual(self.phase_tw.rowCount(), 6)
        self.assertEqual(len(self.model.phase_model.phases), 6)
        self.assertEqual(len(self.widget.pattern_widget.phases), 6)
        self.assertEqual(self.phase_tw.currentRow(), 5)

        click_button(self.phase_widget.delete_btn)
        self.assertEqual(self.phase_tw.rowCount(), 5)
        self.assertEqual(len(self.model.phase_model.phases), 5)
        self.assertEqual(len(self.widget.pattern_widget.phases), 5)
        self.assertEqual(self.phase_tw.currentRow(), 4)

        self.phase_widget.select_phase(1)
        click_button(self.phase_widget.delete_btn)
        self.assertEqual(self.phase_tw.rowCount(), 4)
        self.assertEqual(len(self.model.phase_model.phases), 4)
        self.assertEqual(len(self.widget.pattern_widget.phases), 4)
        self.assertEqual(self.phase_tw.currentRow(), 1)

        self.phase_widget.select_phase(0)
        click_button(self.phase_widget.delete_btn)
        self.assertEqual(self.phase_tw.rowCount(), 3)
        self.assertEqual(len(self.model.phase_model.phases), 3)
        self.assertEqual(len(self.widget.pattern_widget.phases), 3)
        self.assertEqual(self.phase_tw.currentRow(), 0)

        click_button(self.phase_widget.delete_btn)
        click_button(self.phase_widget.delete_btn)
        click_button(self.phase_widget.delete_btn)
        self.assertEqual(self.phase_tw.rowCount(), 0)
        self.assertEqual(len(self.model.phase_model.phases), 0)
        self.assertEqual(len(self.widget.pattern_widget.phases), 0)
        self.assertEqual(self.phase_tw.currentRow(), -1)

        click_button(self.phase_widget.delete_btn)
        self.assertEqual(self.phase_tw.rowCount(), 0)
        self.assertEqual(len(self.model.phase_model.phases), 0)
        self.assertEqual(len(self.widget.pattern_widget.phases), 0)
        self.assertEqual(self.phase_tw.currentRow(), -1)

    def test_automatic_deleting_phases(self):
        self.load_phases()
        self.load_phases()

        self.assertEqual(self.phase_tw.rowCount(), 12)
        self.assertEqual(len(self.model.phase_model.phases), 12)
        self.assertEqual(len(self.widget.pattern_widget.phases), 12)

        click_button(self.phase_widget.clear_btn)
        self.assertEqual(self.phase_tw.rowCount(), 0)
        self.assertEqual(len(self.model.phase_model.phases), 0)
        self.assertEqual(len(self.widget.pattern_widget.phases), 0)
        self.assertEqual(self.phase_tw.currentRow(), -1)

        multiplier = 1
        for dummy_index in range(multiplier):
            self.load_phases()

        self.assertEqual(self.phase_tw.rowCount(), multiplier * 6)

        click_button(self.phase_widget.clear_btn)
        self.assertEqual(self.phase_tw.rowCount(), 0)
        self.assertEqual(len(self.model.phase_model.phases), 0)
        self.assertEqual(len(self.widget.pattern_widget.phases), 0)
        self.assertEqual(self.phase_tw.currentRow(), -1)

    def test_pressure_step_spinbox_changes_pressure_spinboxes(self):
        self.load_phases()
        for ind in range(6):
            self.assertEqual(self.phase_widget.pressure_sbs[ind].singleStep(), 1)

        new_step = 5
        self.phase_widget.pressure_step_msb.setValue(new_step)
        for ind in range(6):
            self.assertEqual(self.phase_widget.pressure_sbs[ind].singleStep(), new_step)

    def test_temperature_step_spinbox_changes_temperature_spinboxes(self):
        self.load_phases()
        for ind in range(6):
            self.assertEqual(self.phase_widget.temperature_sbs[ind].singleStep(), 100)

        new_step = 5
        self.phase_widget.temperature_step_msb.setValue(new_step)
        for ind in range(6):
            self.assertEqual(
                self.phase_widget.temperature_sbs[ind].singleStep(), new_step
            )

    def test_pressure_change(self):
        self.load_phases()
        click_checkbox(self.phase_widget.apply_to_all_cb)

        pressure = 200
        for ind in [0, 1, 3]:
            self.phase_widget.pressure_sbs[ind].setValue(pressure)
            self.assertEqual(
                self.model.phase_model.phases[ind].params["pressure"], pressure
            )
        self.assertEqual(self.model.phase_model.phases[2].params["pressure"], 0)

    def test_temperature_change(self):
        self.load_phases()
        self.model.phase_model.set_pressure(2, 100)  # otherwise the reflections go none

        temperature = 1500

        for ind in range(len(self.model.phase_model.phases)):
            phase = self.model.phase_model.phases[ind]
            temperature += ind

            self.assertEqual(
                self.phase_widget.temperature_sbs[ind].isEnabled(),
                phase.has_thermal_expansion(),
            )

            if self.phase_widget.temperature_sbs[ind].isEnabled():
                self.phase_widget.temperature_sbs[ind].setValue(temperature)

            if phase.has_thermal_expansion():
                self.assertEqual(phase.params["temperature"], temperature)
                self.assertEqual(
                    self.phase_widget.get_phase_temperature(ind), temperature
                )
            else:
                self.assertEqual(phase.params["temperature"], 298)
                self.assertEqual(self.phase_widget.get_phase_temperature(ind), 298)

    def test_pressure_auto_step_change(self):
        self.load_phases()
        self.widget.phase_pressure_step_msb.setValue(0.5)
        self.widget.phase_pressure_step_msb.stepUp()

        new_pressure_step = self.widget.phase_pressure_step_msb.value()
        self.assertAlmostEqual(new_pressure_step, 1.0, places=5)

        self.widget.phase_pressure_step_msb.stepDown()
        self.widget.phase_pressure_step_msb.stepDown()
        new_pressure_step = self.widget.phase_pressure_step_msb.value()
        self.assertAlmostEqual(new_pressure_step, 0.2, places=5)

    def test_temperature_auto_step_change(self):
        self.load_phases()
        self.widget.phase_temperature_step_msb.setValue(10.0)
        self.widget.phase_temperature_step_msb.stepUp()

        new_pressure_step = self.widget.phase_temperature_step_msb.value()
        self.assertAlmostEqual(new_pressure_step, 20.0, places=5)

        self.widget.phase_temperature_step_msb.stepDown()
        self.widget.phase_temperature_step_msb.stepDown()
        new_pressure_step = self.widget.phase_temperature_step_msb.value()
        self.assertAlmostEqual(new_pressure_step, 5.0, places=5)

    def test_apply_to_all_for_new_added_phase_in_table_widget(self):
        temperature = 1500
        pressure = 200
        self.load_phases()
        self.phase_widget.pressure_sbs[0].setValue(pressure)
        self.phase_widget.temperature_sbs[0].setValue(temperature)
        self.load_phases()

        for ind, phase in enumerate(self.model.phase_model.phases):
            self.assertEqual(phase.params["pressure"], pressure)
            self.assertEqual(self.phase_widget.get_phase_pressure(ind), pressure)

    def test_to_not_show_lines_in_legend(self):
        self.load_phases()
        self.phase_tw.selectRow(1)
        QTest.mouseClick(self.widget.phase_del_btn, QtCore.Qt.LeftButton)
        self.widget.pattern_widget.hide_phase(1)

    def test_save_and_load_phase_lists(self):
        # load some phases
        self.load_phases()
        phase_list_file_name = "phase_list.txt"
        QtWidgets.QFileDialog.getSaveFileName = MagicMock(
            return_value=os.path.join(data_path, phase_list_file_name)
        )
        click_button(self.widget.phase_save_list_btn)
        # make sure that phase list file was saved
        self.assertTrue(os.path.isfile(os.path.join(data_path, phase_list_file_name)))

        old_phase_list_length = self.widget.phase_tw.rowCount()
        old_phase_list_data = [
            [0 for x in range(5)] for y in range(old_phase_list_length)
        ]
        for row in range(self.widget.phase_tw.rowCount()):
            old_phase_list_data[row][2] = self.phase_tw.item(row, 2).text()
            old_phase_list_data[row][3] = self.phase_widget.pressure_sbs[row].text()
            old_phase_list_data[row][4] = self.phase_widget.temperature_sbs[row].text()

        # clear and load the saved list to make sure all phases have been loaded
        click_button(self.widget.phase_clear_btn)
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(data_path, phase_list_file_name)
        )
        click_button(self.widget.phase_load_list_btn)

        self.assertEqual(self.widget.phase_tw.rowCount(), old_phase_list_length)

        for row in range(self.widget.phase_tw.rowCount()):
            self.assertEqual(
                self.phase_tw.item(row, 2).text(), old_phase_list_data[row][2]
            )
            self.assertEqual(
                self.phase_widget.pressure_sbs[row].text(), old_phase_list_data[row][3]
            )
            self.assertEqual(
                self.phase_widget.temperature_sbs[row].text(),
                old_phase_list_data[row][4],
            )

        # delete phase list file
        os.remove(os.path.join(data_path, phase_list_file_name))

    def test_bulk_change_visibility_of_phases(self):
        self.load_phases()
        for cb in self.phase_widget.phase_show_cbs:
            self.assertTrue(cb.isChecked())

        self.controller.phase_tw_header_section_clicked(0)
        for cb in self.phase_widget.phase_show_cbs:
            self.assertFalse(cb.isChecked())

        click_checkbox(self.phase_widget.phase_show_cbs[1])
        self.controller.phase_tw_header_section_clicked(0)
        for ind, cb in enumerate(self.phase_widget.phase_show_cbs):
            self.assertFalse(cb.isChecked())

        self.controller.phase_tw_header_section_clicked(0)
        for ind, cb in enumerate(self.phase_widget.phase_show_cbs):
            self.assertTrue(cb.isChecked())

    def test_change_phase_color(self):
        """
        Test that phase color is changed in phase table widget, phase legend and phase line. We assign the color from
        the second color button to the fourth phase.
        """
        self.load_phases()
        new_color = (
            self.phase_widget.phase_color_btns[1].palette().color(QtGui.QPalette.Button)
        )
        QtWidgets.QColorDialog.getColor = MagicMock(return_value=new_color)
        click_button(self.phase_widget.phase_color_btns[3])

        self.assertEqual(
            self.phase_widget.phase_color_btns[3]
            .palette()
            .color(QtGui.QPalette.Button),
            new_color,
        )

        self.assertEqual(
            self.model.phase_model.phase_colors[3],
            (new_color.red(), new_color.green(), new_color.blue()),
        )

        phase_line_color = self.widget.pattern_widget.phases[3].pen.color()
        phase_legend_color = self.widget.pattern_widget.phases_legend.legendItems[3][
            1
        ].opts["color"]
        self.assertEqual(phase_line_color, new_color)
        self.assertEqual(
            phase_legend_color, (new_color.red(), new_color.green(), new_color.blue())
        )

    def load_phases(self):
        self.load_phase("ar.jcpds")
        self.load_phase("ag.jcpds")
        self.load_phase("au_Anderson.jcpds")
        self.load_phase("mo.jcpds")
        self.load_phase("pt.jcpds")
        self.load_phase("re.jcpds")

    def load_phase(self, filename):
        self.model.phase_model.add_jcpds(os.path.join(jcpds_path, filename))


def test_save_phaselist(qapp, tmp_path):
    integration_widget = IntegrationWidget()
    model = DioptasModel()
    model.calibration_model.is_calibrated = True
    model.calibration_model.pattern_geometry.wavelength = 0.31e-10
    model.calibration_model.integrate_1d = MagicMock(
        return_value=(model.calibration_model.tth, model.calibration_model.int)
    )

    phase_controller = PhaseController(integration_widget, model)

    model.phase_model.add_jcpds(os.path.join(jcpds_path, "ar.jcpds"))
    model.phase_model.add_jcpds(os.path.join(jcpds_path, "ag.jcpds"))

    QtWidgets.QFileDialog.getSaveFileName = MagicMock(
        return_value=tmp_path / "test.txt"
    )
    phase_controller.save_btn_clicked_callback()
    phase_controller.clear_phases()

    assert len(model.phase_model.phases) == 0

    QtWidgets.QFileDialog.getOpenFileName = MagicMock(
        return_value=tmp_path / "test.txt"
    )

    phase_controller.load_list_btn_clicked_callback()

    assert len(model.phase_model.phases) == 2


def test_save_phaselist_with_german_locale(qapp, tmp_path):
    QtCore.QLocale.setDefault(QtCore.QLocale(QtCore.QLocale.German))
    test_save_phaselist(qapp, tmp_path)



# SPDX-License-Identifier: MIT

import os
import numpy as np

from ...model.util.HelperModule import get_partial_index, FileNameIterator, get_partial_value

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, '../data', 'FileIterator')


def test_get_partial_index():
    data = np.arange(0, 10, 1)
    value = 2.5
    assert get_partial_index(data, value) == 2.5
    assert get_partial_index(data, data[4]) == 4


def test_get_partial_index_out_of_range():
    data = np.arange(0, 10, 1)
    assert get_partial_index(data, -1) is None
    assert get_partial_index(data, 11) is None


def test_get_partial_value():
    data = np.arange(2, 11, 2)
    assert get_partial_value(data, 1.5) == 5
    assert get_partial_value(data, 0.3) == 2 + 0.3 * 2


def test_get_partial_value_out_of_range():
    data = np.arange(2, 11, 2)
    assert get_partial_value(data, -1) is None
    assert get_partial_value(data, 10) is None


def test_get_next_filename():
    filename = os.path.join(data_path, "dummy1_1.txt")
    file_iterator = FileNameIterator(filename)
    new_filename = file_iterator.get_next_filename(1, filename)

    assert new_filename == os.path.join(data_path, 'dummy1_2.txt')


def test_get_next_filename_with_pos():
    filename = os.path.join(data_path, "dummy1_1.txt")
    file_iterator = FileNameIterator()
    new_filename = file_iterator.get_next_filename(1, filename, pos=1)

    assert new_filename == os.path.join(data_path, 'dummy2_1.txt')


def test_get_previous_filename_with_pos():
    filename = os.path.join(data_path, "dummy2_1.txt")
    file_iterator = FileNameIterator()
    new_filename = file_iterator.get_previous_filename(1, filename, pos=1)

    assert new_filename == os.path.join(data_path, 'dummy1_1.txt')


def test_get_next_folder():
    filename = os.path.join(data_path, "run1", "dummy_1.txt")
    file_iterator = FileNameIterator()
    new_filename = file_iterator.get_next_folder(filename)
    assert new_filename == os.path.join(data_path, 'run2', "dummy_1.txt")


def test_get_previous_folder():
    filename = os.path.join(data_path, "run2", "dummy_1.txt")
    file_iterator = FileNameIterator()
    new_filename = file_iterator.get_previous_folder(filename)
    assert new_filename == os.path.join(data_path, 'run1', "dummy_1.txt")


def test_get_next_folder_mec():
    filename = os.path.join(data_path, "run1", "run_1_evt_2.0.txt")
    file_iterator = FileNameIterator()
    new_filename = file_iterator.get_next_folder(filename, mec_mode=True)
    assert new_filename == os.path.join(data_path, 'run2', "run_2_evt_2.0.txt")


def test_get_previous_folder_mec():
    filename = os.path.join(data_path, "run2", "run_2_evt_2.0.txt")
    file_iterator = FileNameIterator()
    new_filename = file_iterator.get_previous_folder(filename, mec_mode=True)
    assert new_filename == os.path.join(data_path, 'run1', "run_1_evt_2.0.txt")



# SPDX-License-Identifier: MIT

from qtpy import QtWidgets, QtCore
from pyqtgraph import GraphicsLayoutWidget
from dioptas.widgets.plot_widgets import PatternWidget
from dioptas.widgets.plot_widgets.ImgWidget import IntegrationImgWidget

from .integration.CustomWidgets import MouseUnitCurrentAndClickedWidget
from .CustomWidgets import SaveIconButton, CheckableFlatButton, HorizontalSpacerItem, CleanLooksComboBox


class MapWidget(QtWidgets.QWidget):
    """
    Defines the main structure of the Map widget, which is separated into several Parts
    """

    def __init__(self, *args, **kwargs):
        super(MapWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()

    def create_widgets(self):
        self.map_pg_layout = GraphicsLayoutWidget()
        self.map_plot_widget = IntegrationImgWidget(
            self.map_pg_layout, orientation="horizontal"
        )
        self.map_image_frame = MapImageFrame(self.map_pg_layout)

        self.img_pg_layout = GraphicsLayoutWidget()
        self.img_plot_widget = IntegrationImgWidget(
            self.img_pg_layout, orientation="horizontal", padding=0
        )
        self.img_autoscale_btn = CheckableFlatButton("A", self.img_pg_layout)
        self.img_autoscale_btn.setChecked(True)
        self.img_autoscale_btn.setToolTip("Auto-scale image intensity")
        self.img_autoscale_btn.setFixedSize(22, 22)
        self.img_autoscale_btn.raise_()
        self.img_pg_layout.installEventFilter(self)

        self.map_plot_control_widget = (
            MapPlotControlWidget()
        )  # widget below the map image

        self.pattern_pg_layout = GraphicsLayoutWidget()
        self.pattern_plot_widget = PatternWidget(self.pattern_pg_layout)
        self.pattern_plot_widget.show_map_interactive_roi()

        self.pattern_footer_widget = PatternFooterWidget()
        self.pattern_widget = QtWidgets.QWidget()

        self.control_widget = MapControlWidget()

    def eventFilter(self, obj, event):
        if obj is self.img_pg_layout and event.type() == QtCore.QEvent.Resize:
            self.img_autoscale_btn.move(
                self.img_pg_layout.width() - self.img_autoscale_btn.width() - 5,
                self.img_pg_layout.height() - self.img_autoscale_btn.height() - 5,
            )
        return super().eventFilter(obj, event)

    def create_layout(self):
        self._outer_layout = TightHBoxLayout()
        self._left_layout = TightVBoxLayout()
        self._right_layout = TightVBoxLayout()
        self._upper_right_layout = TightHBoxLayout()

        self._left_widget = QtWidgets.QWidget()
        self._left_widget.setLayout(self._left_layout)
        self._left_layout.addWidget(self.map_image_frame)
        self._left_layout.addWidget(self.map_plot_control_widget)

        self.upper_right_splitter = QtWidgets.QSplitter()
        self.upper_right_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.upper_right_splitter.addWidget(self.img_pg_layout)
        self.upper_right_splitter.addWidget(self.control_widget)

        self._lower_right_layout = TightVBoxLayout()
        self._lower_right_layout.addWidget(self.pattern_pg_layout)
        self._lower_right_layout.addWidget(self.pattern_footer_widget)
        self.pattern_widget.setLayout(self._lower_right_layout)

        self.vertical_splitter = QtWidgets.QSplitter(self)
        self.vertical_splitter.setOrientation(QtCore.Qt.Vertical)
        self.vertical_splitter.addWidget(self.upper_right_splitter)
        self.vertical_splitter.addWidget(self.pattern_widget)

        self.horizontal_splitter = QtWidgets.QSplitter()
        self.horizontal_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.horizontal_splitter.addWidget(self._left_widget)
        self.horizontal_splitter.addWidget(self.vertical_splitter)

        self._outer_layout.addWidget(self.horizontal_splitter)

        self.setLayout(self._outer_layout)


class TightHBoxLayout(QtWidgets.QHBoxLayout):
    def __init__(self, *args, **kwargs):
        super(TightHBoxLayout, self).__init__(*args, **kwargs)
        self.setContentsMargins(0, 0, 0, 0)
        self.setSpacing(0)


class TightVBoxLayout(QtWidgets.QVBoxLayout):
    def __init__(self, *args, **kwargs):
        super(TightVBoxLayout, self).__init__(*args, **kwargs)
        self.setContentsMargins(0, 0, 0, 0)
        self.setSpacing(0)


class MapImageFrame(QtWidgets.QWidget):
    """Frame wrapping the map plot with a black control bar matching integration widget style."""

    def __init__(self, map_pg_layout: GraphicsLayoutWidget, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.frame = QtWidgets.QWidget()
        self.frame.setObjectName('map_image_frame')

        self.map_pg_layout = map_pg_layout

        self.smooth_btn = CheckableFlatButton("Smooth")
        self.smooth_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.smooth_slider.setRange(2, 10)
        self.smooth_slider.setValue(5)
        self.smooth_slider.setMaximumWidth(160)
        self.smooth_slider.setVisible(False)
        self.smooth_label = QtWidgets.QLabel("5")
        self.smooth_label.setFixedWidth(20)
        self.smooth_label.setVisible(False)

        self.contour_btn = CheckableFlatButton("Contours")
        self.contour_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.contour_slider.setRange(3, 20)
        self.contour_slider.setValue(5)
        self.contour_slider.setMaximumWidth(160)
        self.contour_slider.setVisible(False)
        self.contour_label = QtWidgets.QLabel("5")
        self.contour_label.setFixedWidth(20)
        self.contour_label.setVisible(False)

        self._control_layout = QtWidgets.QHBoxLayout()
        self._control_layout.setContentsMargins(6, 2, 6, 2)
        self._control_layout.setSpacing(6)
        self._control_layout.addWidget(self.smooth_btn)
        self._control_layout.addWidget(self.smooth_slider)
        self._control_layout.addWidget(self.smooth_label)
        self._control_layout.addSpacerItem(HorizontalSpacerItem())
        self._control_layout.addWidget(self.contour_label)
        self._control_layout.addWidget(self.contour_slider)
        self._control_layout.addWidget(self.contour_btn)

        self._frame_layout = QtWidgets.QVBoxLayout()
        self._frame_layout.setContentsMargins(0, 0, 0, 0)
        self._frame_layout.setSpacing(0)
        self._frame_layout.addWidget(self.map_pg_layout)
        self._frame_layout.addLayout(self._control_layout)
        self.frame.setLayout(self._frame_layout)

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.frame)
        self.setLayout(self._layout)

        self.setStyleSheet("""
            #map_image_frame {
                background: black;
            }
            #map_image_frame QPushButton {
                padding-top: 2px;
                padding-bottom: 2px;
                padding-left: 5px;
                padding-right: 5px;
            }
            #map_image_frame QSlider {
                background: transparent;
            }
            #map_image_frame QLabel {
                background: transparent;
            }
        """)


class MapControlWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(MapControlWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()

    def create_widgets(self):
        self.load_btn = QtWidgets.QPushButton("Load")
        self.file_list = QtWidgets.QListWidget()
        self.reintegrate_cb = QtWidgets.QCheckBox("Reintegrate")
        self.reintegrate_cb.setToolTip(
            "Re-integrate the pattern when selecting a new image.\nApplying all current integration settings to the new pattern."
        )
        self.reintegrate_cb.setChecked(False)

    def create_layout(self):
        self._outer_layout = TightVBoxLayout()
        self._outer_layout.setContentsMargins(0, 0, 0, 0)
        self._outer_layout.setSpacing(5)

        self._outer_layout.addWidget(self.load_btn)
        self._outer_layout.addWidget(self.file_list)
        self._outer_layout.addWidget(self.reintegrate_cb)

        self.setLayout(self._outer_layout)

    def style_widgets(self):
        pass


class MapPlotControlWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(MapPlotControlWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()

    def create_widgets(self):
        self.save_map_btn = SaveIconButton()
        self.map_dimension_cb = CleanLooksComboBox()
        self.map_dimension_cb.setMinimumWidth(80)
        self.mouse_x_label = QtWidgets.QLabel("X: ")
        self.mouse_y_label = QtWidgets.QLabel("Y: ")
        self.mouse_int_label = QtWidgets.QLabel("I: ")
        self.filename_label = QtWidgets.QLabel("")

    def create_layout(self):
        self._outer_layout = TightHBoxLayout()

        self._left_layout = TightVBoxLayout()
        self._mouse_pos_layout = TightHBoxLayout()
        self._mouse_pos_layout.addWidget(self.mouse_x_label)
        self._mouse_pos_layout.addWidget(self.mouse_y_label)
        self._mouse_pos_layout.addWidget(self.mouse_int_label)
        self._left_layout.addLayout(self._mouse_pos_layout)
        self._left_layout.addWidget(self.filename_label)
        self._outer_layout.addWidget(self.save_map_btn)
        self._outer_layout.addWidget(QtWidgets.QLabel("Dim: "))
        self._outer_layout.addWidget(self.map_dimension_cb)
        self._outer_layout.addStretch(1)
        self._outer_layout.addLayout(self._left_layout)
        self.setLayout(self._outer_layout)

    def style_widgets(self):
        self._outer_layout.setContentsMargins(6, 3, 0, 0)
        self.mouse_x_label.setFixedWidth(50)
        self.mouse_y_label.setFixedWidth(50)
        self.mouse_int_label.setMinimumWidth(80)
        self.setMinimumHeight(30)


class PatternFooterWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(PatternFooterWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()

    def create_widgets(self):
        self.mouse_unit_widget = MouseUnitCurrentAndClickedWidget()
        self.log_btn = CheckableFlatButton('Log')
        self.sqrt_btn = CheckableFlatButton(u'\u221a')

    def create_layout(self):
        self._outer_layout = TightHBoxLayout()
        self._outer_layout.addWidget(self.mouse_unit_widget)
        self._outer_layout.addStretch(1)
        self._outer_layout.addWidget(self.log_btn)
        self._outer_layout.addWidget(self.sqrt_btn)
        self.setLayout(self._outer_layout)

    def style_widgets(self):
        self._outer_layout.setContentsMargins(6, 3, 6, 0)
        self.setMinimumHeight(30)



# SPDX-License-Identifier: MIT

import os
import pytest
from mock import MagicMock

from ..utility import click_button

from qtpy import QtWidgets, QtGui

from ...widgets.plot_widgets.ImgWidget import MyRectangle

from .test_BatchController_part1 import *


def test_set_range_img(batch_widget, batch_controller, load_proc_data):
    batch_widget.position_widget.step_series_widget.start_txt.setValue(25)
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(50)

    batch_controller.set_range_img()
    assert batch_widget.position_widget.step_series_widget.slider.value() == 25

    batch_widget.position_widget.step_series_widget.start_txt.setValue(5)
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(20)
    batch_controller.set_range_img()
    assert batch_widget.position_widget.step_series_widget.slider.value() == 20


def test_show_img_mouse_position(batch_widget, batch_controller, load_proc_data):
    batch_widget.activate_stack_plot()
    batch_controller.show_img_mouse_position(10, 15)

    assert (
        batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.x_pos_lbl.text()
        == "Img: 15"
    )
    assert (
        batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.y_pos_lbl.text()
        == "2θ: 9.7"
    )
    assert (
        batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.int_lbl.text()
        == "I: 0.1"
    )


def test_load_raw_data(batch_widget, batch_controller, batch_model, load_proc_data):
    files = [
        os.path.join(
            unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00000.nxs"
        ),
        os.path.join(
            unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00001.nxs"
        ),
    ]

    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(return_value=files)
    batch_controller.load_data()

    assert batch_model.data is None
    assert batch_model.binning is None
    assert batch_model.raw_available

    start, stop, step = batch_widget.position_widget.step_raw_widget.get_image_range()
    frame = str(batch_widget.position_widget.step_series_widget.pos_label.text())
    assert stop == 19
    assert start == 0
    assert frame == "Frame(None/20):"

    assert batch_widget.file_view_widget.tree_model.columnCount() == 2
    assert (batch_model.files == files).all()


def test_load_proc_data(
    batch_widget, batch_controller, batch_model, dioptas_model, load_proc_data
):
    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_proc.nxs"
    )
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(return_value=[filename])
    dioptas_model.working_directories["image"] = os.path.join(
        unittest_data_path, "lambda"
    )
    batch_controller.load_data()

    assert batch_model.data is not None
    # self.assertTrue(self.model.batch_model.raw_available)
    assert batch_model.data.shape[0] == 50
    assert batch_model.data.shape[1] == batch_model.binning.shape[0]
    assert batch_model.data.shape[0] == batch_model.n_img
    start = int(str(batch_widget.position_widget.step_series_widget.start_txt.text()))
    stop = int(str(batch_widget.position_widget.step_series_widget.stop_txt.text()))
    frame = str(batch_widget.position_widget.step_series_widget.pos_label.text())
    assert stop == 49
    assert start == 0
    assert frame == "Frame(50/50):"

    assert (
        batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl.text()
        == "Img: 0"
    )
    assert batch_widget.position_widget.step_series_widget.slider.value() == 0

    filename = (
        os.path.join(
            unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00000.nxs"
        )
        .split("/")[-1]
        .split("\\")[-1]
    )
    assert filename in batch_widget.windowTitle()


def test_plot_batch_2d(batch_widget, batch_controller, batch_model, load_proc_data):
    batch_widget.position_widget.step_series_widget.start_txt.setValue(10)
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(40)
    batch_widget.activate_stack_plot()

    batch_controller.plot_batch()
    assert batch_widget.stack_plot_widget.img_view.img_data.shape == (31, 4038)
    assert batch_widget.stack_plot_widget.img_view._max_range
    assert batch_widget.stack_plot_widget.img_view.horizontal_line.value() == 0
    assert batch_widget.stack_plot_widget.img_view.left_axis_cake.range[
        0
    ] == pytest.approx(8.0766, abs=0.1)
    assert batch_widget.stack_plot_widget.img_view.left_axis_cake.range[
        1
    ] == pytest.approx(41.8757, abs=0.1)


def test_plot_batch_3d(batch_widget, batch_controller, batch_model, load_proc_data):
    batch_widget.activate_surface_view()
    batch_widget.position_widget.step_series_widget.start_txt.blockSignals(True)
    batch_widget.position_widget.step_series_widget.stop_txt.blockSignals(True)
    batch_widget.position_widget.step_series_widget.start_txt.setValue(10)
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(40)
    batch_widget.position_widget.step_series_widget.start_txt.blockSignals(False)
    batch_widget.position_widget.step_series_widget.stop_txt.blockSignals(False)
    batch_controller.plot_batch()

    assert batch_widget.position_widget.step_series_widget.step_txt.value() == 1
    assert batch_widget.surface_widget.surface_view.data.shape == (31, 4038)


def test_img_mouse_click(batch_widget, batch_controller, batch_model, load_proc_data):
    # Test only image loading. Waterfall is already tested
    batch_controller.img_mouse_click(10, 15)

    assert batch_widget.surface_widget.surface_view.g_translate == 0
    assert (
        batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl.text()
        == "Img: 15"
    )
    assert batch_widget.position_widget.step_series_widget.slider.value() == 15

    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00001.nxs"
    )
    assert batch_widget.windowTitle() == f"Batch widget. {filename} - 5"


def test_process_waterfall(batch_controller, load_proc_data):
    batch_controller.process_waterfall(5, 7)
    assert batch_controller.rect.rect().left() == 5
    assert batch_controller.rect.rect().bottom() == 7
    assert batch_controller.clicks == 1

    batch_controller.process_waterfall(10, 17)
    assert batch_controller.clicks == 0


def test_plot_waterfall(batch_controller, dioptas_model, load_proc_data):
    batch_controller.rect = MyRectangle(5, 7, 10, 17, QtGui.QColor(255, 0, 0, 150))

    batch_controller.plot_waterfall()

    assert len(dioptas_model.overlay_model.overlays) == 17
    assert (
        dioptas_model.overlay_model.overlays[0].name
        == "testasapo1_1009_00002_m1_part00002.nxs, 4"
    )
    assert dioptas_model.overlay_model.overlays[0]._pattern_x.shape == (10,)


def test_normalize(batch_widget, batch_controller, load_proc_data):
    click_button(batch_widget.control_widget.normalize_btn)



# SPDX-License-Identifier: MIT


# imports for type hinting in PyCharm -- DO NOT DELETE
from ...widgets.integration import IntegrationWidget
from ...model.DioptasModel import DioptasModel


class OptionsController(object):
    """
    IntegrationPatternController handles all the interaction from the IntegrationView with the pattern data.
    It manages the auto integration of image files to  in addition to pattern browsing and changing of units
    (2 Theta, Q, A)
    """

    def __init__(self, widget, dioptas_model):
        """
        :param widget: Reference to an IntegrationWidget
        :param dioptas_model: reference to DioptasModel object

        :type widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        """

        self.integration_widget = widget
        self.options_widget = self.integration_widget.integration_control_widget.integration_options_widget

        self.model = dioptas_model

        self._setup_dioptrin_checkbox()
        self.connect_signals()

    def _setup_dioptrin_checkbox(self):
        import dioptas

        self._dioptrin_available = dioptas._dioptrin_available
        if not self._dioptrin_available:
            self.options_widget.use_dioptrin_cb.setVisible(False)

    def connect_signals(self):
        self.options_widget.correct_solid_angle_cb.stateChanged.connect(self.correct_solid_angle_cb_clicked)
        self.model.configuration_selected.connect(self.update_gui)
        self.model.pattern_changed.connect(self.update_gui)
        self.options_widget.use_dioptrin_cb.toggled.connect(self._use_dioptrin_toggled)

        self.options_widget.cake_azimuth_points_sb.valueChanged.connect(self.cake_azimuth_points_changed)
        self.options_widget.cake_azimuth_min_txt.editingFinished.connect(self.cake_azimuth_range_changed)
        self.options_widget.cake_azimuth_max_txt.editingFinished.connect(self.cake_azimuth_range_changed)

        self.options_widget.oned_full_toggle_btn.toggled.connect(self.oned_full_toggled_btn_changed)
        self.options_widget.cake_full_toggle_btn.toggled.connect(self.cake_full_toggled_btn_changed)
        self.options_widget.oned_azimuth_min_txt.editingFinished.connect(self.oned_azimuth_range_changed)
        self.options_widget.oned_azimuth_max_txt.editingFinished.connect(self.oned_azimuth_range_changed)

    def _use_dioptrin_toggled(self, checked):
        self.model.calibration_model.use_dioptrin = checked
        if checked and self.model.calibration_model.is_calibrated:
            self.model.calibration_model._create_dioptrin_integrator()
        if self.model.calibration_model.is_calibrated:
            self.model.current_configuration.integrate_image_1d()
            if self.model.current_configuration._auto_integrate_cake:
                self.model.current_configuration.integrate_image_2d()

    def correct_solid_angle_cb_clicked(self):
        self.model.current_configuration.correct_solid_angle = self.options_widget.correct_solid_angle_cb.isChecked()

    def update_gui(self):
        self.options_widget.blockSignals(True)

        self.options_widget.correct_solid_angle_cb.blockSignals(True)
        self.options_widget.correct_solid_angle_cb.setChecked(int(self.model.current_configuration.correct_solid_angle))
        self.options_widget.correct_solid_angle_cb.blockSignals(False)

        self.options_widget.use_dioptrin_cb.blockSignals(True)
        self.options_widget.use_dioptrin_cb.setChecked(self.model.calibration_model.use_dioptrin)
        self.options_widget.use_dioptrin_cb.blockSignals(False)

        self.options_widget.bin_count_txt.blockSignals(True)
        self.options_widget.bin_count_txt.setText("{:1.0f}".format(self.model.calibration_model.num_points))
        self.options_widget.bin_count_txt.blockSignals(False)

        self.options_widget.cake_azimuth_points_sb.blockSignals(True)
        self.options_widget.cake_azimuth_points_sb.setValue(self.model.current_configuration.cake_azimuth_points)
        self.options_widget.cake_azimuth_points_sb.blockSignals(False)

        if self.model.current_configuration.cake_azimuth_range is None:
            self.enable_full_cake_range()
        else:
            self.options_widget.cake_azimuth_min_txt.setText(
                '{}'.format(self.model.current_configuration.cake_azimuth_range[0]))
            self.options_widget.cake_azimuth_max_txt.setText(
                '{}'.format(self.model.current_configuration.cake_azimuth_range[1]))
            self.options_widget.blockSignals(False)
            self.disable_full_cake_range()
        self.options_widget.blockSignals(False)

    def cake_azimuth_range_changed(self):
        range_min = float(self.options_widget.cake_azimuth_min_txt.text())
        range_max = float(self.options_widget.cake_azimuth_max_txt.text())
        self.model.current_configuration.cake_azimuth_range = (range_min, range_max)

    def cake_azimuth_points_changed(self):
        self.model.current_configuration.cake_azimuth_points = int(
            self.options_widget.cake_azimuth_points_sb.value())

    def cake_full_toggled_btn_changed(self):
        if self.options_widget.cake_full_toggle_btn.isChecked():
            self.enable_full_cake_range()
            self.model.current_configuration.cake_azimuth_range = None

        elif not self.options_widget.cake_full_toggle_btn.isChecked():
            self.disable_full_cake_range()
            self.cake_azimuth_range_changed()

    def enable_full_cake_range(self):
        self.options_widget.cake_azimuth_min_txt.setDisabled(True)
        self.options_widget.cake_azimuth_max_txt.setDisabled(True)
        self.integration_widget.cake_shift_azimuth_sl.setDisabled(False)

    def disable_full_cake_range(self):
        self.options_widget.cake_azimuth_min_txt.setDisabled(False)
        self.options_widget.cake_azimuth_max_txt.setDisabled(False)
        self.integration_widget.cake_shift_azimuth_sl.setDisabled(True)
        self.integration_widget.cake_shift_azimuth_sl.setValue(0)

    def oned_azimuth_range_changed(self):
        range_min = float(self.options_widget.oned_azimuth_min_txt.text())
        range_max = float(self.options_widget.oned_azimuth_max_txt.text())
        self.model.current_configuration.oned_azimuth_range = (range_min, range_max)

    def oned_full_toggled_btn_changed(self):
        if self.options_widget.oned_full_toggle_btn.isChecked():
            self.enable_full_oned_range()
            self.model.current_configuration.oned_azimuth_range = None

        elif not self.options_widget.oned_full_toggle_btn.isChecked():
            self.disable_full_oned_range()
            self.oned_azimuth_range_changed()

    def enable_full_oned_range(self):
        self.options_widget.oned_azimuth_min_txt.setDisabled(True)
        self.options_widget.oned_azimuth_max_txt.setDisabled(True)

    def disable_full_oned_range(self):
        self.options_widget.oned_azimuth_min_txt.setDisabled(False)
        self.options_widget.oned_azimuth_max_txt.setDisabled(False)



# SPDX-License-Identifier: MIT

from ..utility import QtTest
from mock import MagicMock, patch
import pytest

from ...model.DioptasModel import DioptasModel
from ...widgets.integration import IntegrationWidget
from ...controller.integration.EpicsController import EpicsController

from qtpy import QtWidgets

# Check if epics module is available
try:
    import epics
    EPICS_AVAILABLE = True
except ImportError:
    EPICS_AVAILABLE = False


class TestEpicsController(QtTest):

    def setUp(self):
        # Mock the functions which will block the unittest for some reason...
        # Use patch to ensure proper cleanup
        self.processEvents_patcher = patch.object(QtWidgets.QApplication, 'processEvents', MagicMock())
        self.setValue_patcher = patch.object(QtWidgets.QProgressDialog, 'setValue', MagicMock())
        self.processEvents_patcher.start()
        self.setValue_patcher.start()

        self.integration_widget = IntegrationWidget()
        self.move_widget = self.integration_widget.move_widget
        self.setup_widget = self.integration_widget.move_widget.motors_setup_widget
        self.model = DioptasModel()
        self.epics_controller = EpicsController(self.integration_widget, self.model)

    def tearDown(self):
        # Stop the patchers to restore original behavior
        self.processEvents_patcher.stop()
        self.setValue_patcher.stop()

        # Clean up widgets
        self.integration_widget.close()
        self.integration_widget.deleteLater()
        del self.integration_widget
        del self.model
        del self.epics_controller

    @pytest.mark.skipif(not EPICS_AVAILABLE, reason="epics module not installed")
    @patch('epics.caget', return_value=12.03)
    def test_update_motor_position(self, caget):
        self.epics_controller.update_current_motor_position()

        self.assertEqual(self.move_widget.hor_lbl.text(), '12.03')
        self.assertEqual(self.move_widget.ver_lbl.text(), '12.03')
        self.assertEqual(self.move_widget.focus_lbl.text(), '12.03')
        self.assertEqual(self.move_widget.omega_lbl.text(), '12.03')

    def test_update_image_position(self):
        self.epics_controller.update_image_position()

        self.assertEqual(self.move_widget.img_hor_lbl.text(), '')
        self.assertEqual(self.move_widget.img_ver_lbl.text(), '')
        self.assertEqual(self.move_widget.img_focus_lbl.text(), '')
        self.assertEqual(self.move_widget.img_omega_lbl.text(), '')

        self.model.img_model.motors_info['Horizontal'] = 0.1
        self.model.img_model.motors_info['Vertical'] = 0.2
        self.model.img_model.motors_info['Focus'] = 0.3
        self.model.img_model.motors_info['Omega'] = 0.4

        self.epics_controller.update_image_position()
        self.assertEqual(str(self.move_widget.img_hor_lbl.text()), '0.100')
        self.assertEqual(str(self.move_widget.img_ver_lbl.text()), '0.200')
        self.assertEqual(str(self.move_widget.img_focus_lbl.text()), '0.300')
        self.assertEqual(str(self.move_widget.img_omega_lbl.text()), '0.400')

    @pytest.mark.skipif(not EPICS_AVAILABLE, reason="epics module not installed")
    @patch('epics.caput')
    @patch('epics.caget', return_value=0.0)
    def test_move_stage(self, caget, caput):
        self.model.img_model.motors_info['Horizontal'] = 0.1
        self.model.img_model.motors_info['Vertical'] = 0.02
        self.model.img_model.motors_info['Focus'] = 0.05
        self.model.img_model.motors_info['Omega'] = 90

        self.epics_controller.update_image_position()

        self.move_widget.move_hor_cb.setChecked(True)
        self.move_widget.move_ver_cb.setChecked(True)
        self.move_widget.move_focus_cb.setChecked(True)

        self.epics_controller.move_stage()

        caput.assert_any_call('13IDD:m81.VAL', 0.1)
        caput.assert_any_call('13IDD:m83.VAL', 0.02)
        caput.assert_any_call('13IDD:m82.VAL', 0.05)



# SPDX-License-Identifier: MIT

import pytest
import numpy as np

from dioptas.controller.integration.overlay.OverlayInPatternController import (
    OverlayInPatternController,
)
from dioptas.widgets.integration import IntegrationWidget
from dioptas.widgets.plot_widgets import PatternWidget
from ...model.OverlayModel import OverlayModel
from ...model.util.HelperModule import rgb_to_hex


@pytest.fixture
def overlay_model() -> OverlayModel:
    return OverlayModel()


@pytest.fixture
def overlay_in_pattern_controller(
    pattern_widget, overlay_model
) -> OverlayInPatternController:
    return OverlayInPatternController(pattern_widget, overlay_model)


@pytest.fixture
def pattern_widget(integration_widget: IntegrationWidget) -> PatternWidget:
    return integration_widget.pattern_widget


def test_adding_overlays(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    assert pattern_widget.legend.numItems == 1

    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test2")
    assert len(pattern_widget.overlays) == 2
    assert pattern_widget.legend.numItems == 3

    assert pattern_widget.legend.legendItems[1][1].text == "test1"
    assert pattern_widget.legend.legendItems[2][1].text == "test2"


def test_removing_overlays(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test2")

    overlay_model.remove_overlay(0)
    assert len(pattern_widget.overlays) == 1
    assert len(pattern_widget.legend.legendItems) == 2
    assert pattern_widget.legend.numItems == 2
    assert pattern_widget.legend.legendItems[1][1].text == "test2"


def test_rename_overlay(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test2")

    overlay_model.set_overlay_name(0, "test3")
    assert pattern_widget.legend.legendItems[1][1].text == "test3"


def test_set_overlay_color(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test2")

    new_color = rgb_to_hex((120, 100, 10))
    overlay_model.set_overlay_color(0, new_color)
    assert pattern_widget.legend.legendItems[1][1].opts["color"] == new_color
    assert pattern_widget.overlays[0].opts["pen"].color().name() == new_color


def test_move_overlay_up(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(11), np.arange(11), name="test2")
    overlay_model.add_overlay(np.arange(12), np.arange(12), name="test3")

    overlay_model.move_overlay_up(2)
    assert pattern_widget.legend.legendItems[1][1].text == "test1"
    assert pattern_widget.legend.legendItems[2][1].text == "test3"
    assert pattern_widget.legend.legendItems[3][1].text == "test2"

    assert np.array_equal(pattern_widget.overlays[0].xData, np.arange(10))
    assert np.array_equal(pattern_widget.overlays[1].xData, np.arange(12))
    assert np.array_equal(pattern_widget.overlays[2].xData, np.arange(11))


def test_move_overlay_down(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(11), np.arange(11), name="test2")
    overlay_model.add_overlay(np.arange(12), np.arange(12), name="test3")

    overlay_model.move_overlay_down(0)
    assert pattern_widget.legend.legendItems[1][1].text == "test2"
    assert pattern_widget.legend.legendItems[2][1].text == "test1"
    assert pattern_widget.legend.legendItems[3][1].text == "test3"

    assert np.array_equal(pattern_widget.overlays[0].xData, np.arange(11))
    assert np.array_equal(pattern_widget.overlays[1].xData, np.arange(10))
    assert np.array_equal(pattern_widget.overlays[2].xData, np.arange(12))


def test_hide_overlay(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(11), np.arange(11), name="test2")
    overlay_model.add_overlay(np.arange(12), np.arange(12), name="test3")

    overlay_model.set_overlay_visible(0, False)

    assert not pattern_widget.overlays[0] in pattern_widget.pattern_plot.items
    assert pattern_widget.overlays[1] in pattern_widget.pattern_plot.items
    assert pattern_widget.overlays[2] in pattern_widget.pattern_plot.items


def test_change_scaling(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(11), np.arange(11), name="test2")

    overlay_model.set_overlay_scaling(0, 0.5)
    assert np.array_equal(pattern_widget.overlays[0].yData, np.arange(10) * 0.5)

    overlay_model.set_overlay_scaling(1, 2)
    assert np.array_equal(pattern_widget.overlays[1].yData, np.arange(11) * 2)


def test_change_offset(
    overlay_in_pattern_controller: OverlayInPatternController,
    pattern_widget: PatternWidget,
    overlay_model: OverlayModel,
):
    overlay_model.add_overlay(np.arange(10), np.arange(10), name="test1")
    overlay_model.add_overlay(np.arange(11), np.arange(11), name="test2")

    overlay_model.set_overlay_offset(0, 0.5)
    assert np.array_equal(pattern_widget.overlays[0].yData, np.arange(10) + 0.5)

    overlay_model.set_overlay_offset(1, 2)
    assert np.array_equal(pattern_widget.overlays[1].yData, np.arange(11) + 2)



import os

from ..utility import QtTest, click_button

from ...widgets.integration import IntegrationWidget
from ...controller.integration.BatchController import BatchController
from ...model.DioptasModel import DioptasModel
from dioptas.controller.integration.phase.PhaseController import PhaseController
from dioptas import resources_path

unittest_data_path = os.path.join(os.path.dirname(__file__), '../data')
jcpds_path = os.path.join(unittest_data_path, 'jcpds')


class SurfaceView(QtTest):
    def setUp(self):
        self.working_dir = {'image': ''}

        self.widget = IntegrationWidget()
        self.set_stylesheet()
        self.model = DioptasModel()

        self.controller = BatchController(
            widget=self.widget,
            dioptas_model=self.model)

        self.phase_controller = PhaseController(self.widget, self.model)

        # Load existing proc+raw data
        filename = os.path.join(unittest_data_path, 'lambda', 'testasapo1_1009_00002_proc.nxs')
        self.model.batch_model.load_proc_data(filename)
        raw_files = self.model.batch_model.files
        raw_files = [os.path.join(os.path.dirname(filename), os.path.basename(f)) for f in raw_files]
        self.model.batch_model.set_image_files(raw_files)
        self.widget.batch_widget.position_widget.step_series_widget.stop_txt.setValue(self.model.batch_model.n_img - 1)

    def set_stylesheet(self):
        with open(os.path.join(resources_path, "style", "stylesheet.qss")) as file:
            stylesheet = file.read()
            self.widget.setStyleSheet(stylesheet)

    def test_surface_widget(self):
        self.widget.batch_widget.show()
        click_button(self.widget.batch_widget.mode_widget.view_3d_btn)
        self.app.processEvents()

    def test_batch_widget_design(self):
        self.widget.batch_widget.raise_widget()
        self.app.processEvents()



# SPDX-License-Identifier: MIT
"""Shared integration utilities for MapModel and BatchModel.

Provides frame generators and helpers that deduplicate the batch
integration logic shared between map and batch processing.
"""
from __future__ import annotations

import os

import numpy as np


def detect_fast_path(img_model) -> bool:
    """Check whether the img_model can bypass transformations/corrections.

    When True, raw frames from HDF5 bitshuffle parallel decompression can
    be fed directly to the integrator without applying the frame pipeline.
    """
    return (
        not img_model.img_transformations
        and img_model._background_data is None
        and not img_model._img_corrections.has_items()
        and img_model._factor == 1
    )


def try_open_bitshuffle_hdf5(filepath):
    """Try to open *filepath* as a bitshuffle-compressed HDF5.

    Returns an ``Hdf5Image`` instance if the file is a bitshuffle HDF5,
    otherwise returns ``None`` (closing any opened handle).
    """
    from dioptas.model.loader.hdf5Loader import Hdf5Image

    try:
        loader = Hdf5Image(filepath)
    except Exception:
        return None

    if loader._is_bitshuffle:
        return loader

    loader.f.close()
    return None


def _apply_frame(img_model, raw_frame, fast_path: bool):
    """Apply the img_model frame pipeline if *fast_path* is False."""
    if fast_path:
        return raw_frame
    return img_model._apply_frame_pipeline(raw_frame)


def _open_file_frames(filepath, img_model, fast_path, *, load=False):
    """Open *filepath* and return a generator yielding processed frames.

    Tries bitshuffle HDF5 parallel decompression first for fast frame
    iteration; falls back to ``img_model.load`` +
    ``img_model.load_series_img`` per frame.

    When *load* is True, ``img_model.load(filepath)`` is always called
    so that img_model state (img_data, series_max, etc.) reflects the
    file.  This is needed when callers rely on img_model metadata
    (e.g. for mask dimension).
    """
    hdf5_loader = try_open_bitshuffle_hdf5(filepath)
    if hdf5_loader is not None:
        if load:
            img_model.load(filepath)

        def bitshuffle_gen():
            for raw in hdf5_loader.gen_frames():
                yield _apply_frame(img_model, raw, fast_path)
        return bitshuffle_gen()

    # Fallback: always loads through img_model
    img_model.load(filepath)

    def fallback_gen():
        for i in range(img_model.series_max):
            img_model.load_series_img(i + 1)
            yield img_model.get_img_data_float64()
    return fallback_gen()


def iter_frames_sequential(img_model, filepaths, *, img_shape, abort_check=None, on_frame=None):
    """Yield image frames from *filepaths* sequentially.

    Used by MapModel for map integration.  Tries bitshuffle HDF5 parallel
    decompression first, falls back to standard loading via ``img_model``.

    Args:
        img_model: ImgModel instance for loading/processing frames.
        filepaths: list of file paths to iterate over.
        img_shape: expected image shape — raises ValueError on mismatch.
        abort_check: callable returning True when processing should stop.
        on_frame: ``on_frame(filepath, frame_index)`` called before each
            frame is yielded; use this to collect per-frame metadata.
    """
    fast_path = detect_fast_path(img_model)

    for filepath in filepaths:
        if abort_check and abort_check():
            return

        frame_iter = _open_file_frames(filepath, img_model, fast_path)

        for frame_ind, frame in enumerate(frame_iter):
            if abort_check and abort_check():
                return
            if frame_ind == 0 and frame.shape != img_shape:
                raise ValueError(
                    f"Image '{os.path.basename(filepath)}' has shape "
                    f"{frame.shape}, expected {img_shape}"
                )
            if on_frame:
                on_frame(filepath, frame_ind)
            yield frame


def iter_frames_indexed(img_model, files, source, indices, *, mask_model=None, abort_check=None):
    """Yield image frames at specific ``(file_index, frame_pos)`` positions.

    Used by BatchModel for batch integration with ``pos_map`` indexing.
    When a bitshuffle HDF5 is detected, uses parallel decompression with
    forward-seeking to the requested frame position.

    Args:
        img_model: ImgModel instance for loading/processing frames.
        files: array of file paths (indexed by ``file_index``).
        source: position-map array of ``(file_index, frame_pos)`` entries.
        indices: list of indices into *source* to process.
        mask_model: optional MaskModel — its dimension is updated on file change.
        abort_check: callable returning True when processing should stop.
    """
    fast_path = detect_fast_path(img_model)
    current_file = ""
    frame_iter = None
    next_frame_pos = 0

    for index in indices:
        if abort_check and abort_check():
            return
        file_index, pos = source[index]
        if file_index != current_file:
            current_file = file_index
            frame_iter = _open_file_frames(
                files[file_index], img_model, fast_path, load=True,
            )
            if mask_model is not None:
                mask_model.set_dimension(img_model.img_data.shape)
            next_frame_pos = 0

        # Advance to requested position
        frame = None
        while next_frame_pos <= pos:
            frame = next(frame_iter)
            next_frame_pos += 1
        yield frame


def convert_tth_to_d(tth_array, wavelength):
    """Convert two-theta (degrees) to d-spacing (Angstrom)."""
    return wavelength / (2 * np.sin(tth_array / 360 * np.pi)) * 1e10



# SPDX-License-Identifier: MIT
"""Test NormalizedImageItem pyqtgraph's GraphicObject"""

import numpy as np

import pytest
from pyqtgraph import GraphicsLayoutWidget
from qtpy import QtCore
from qtpy.QtTest import QSignalSpy, QTest

from dioptas.widgets.plot_widgets.NormalizedImageItem import NormalizedImageItem

NORMALIZATIONS = tuple(NormalizedImageItem._NORMALIZATIONS.keys())


@pytest.fixture
def normalizedImageItem(qapp, qWidgetFactory):
    """Fixture providing a NormalizedImageItem displayed in a GraphicsLayoutWidget"""
    widget = qWidgetFactory(GraphicsLayoutWidget)
    viewbox = widget.addViewBox(row=1, col=1)

    item = NormalizedImageItem()
    viewbox.addItem(item)

    yield item
    qapp.processEvents()


def testDefaultItem(normalizedImageItem):
    """Test NormalizedImageItem default values"""
    normalization = normalizedImageItem.getNormalization()
    assert normalization == "linear"
    data = normalizedImageItem.getData(copy=False)
    assert data is None
    levels = normalizedImageItem.getLevels()
    assert levels is None


@pytest.mark.parametrize("normalization", NORMALIZATIONS)
def testSetLevels(normalizedImageItem, normalization):
    """Test setLevels with different normalizations"""
    normalizedImageItem.setNormalization(normalization)
    assert normalizedImageItem.getNormalization() == normalization
    assert normalizedImageItem.getLevels() is None

    normalizedImageItem.setLevels((1, 10))
    levels = normalizedImageItem.getLevels()
    assert np.allclose(levels, (1, 10))


@pytest.mark.parametrize("normalization", NORMALIZATIONS)
def testSetImage(qapp, normalizedImageItem, normalization):
    """Test setImage with different normalizations"""
    normalizedImageItem.setNormalization(normalization)

    ref_image = np.arange(10000, dtype=np.float32).reshape(100, 100)
    min_max = ref_image.min(), ref_image.max()

    normalizedImageItem.setImage(ref_image)
    qapp.processEvents()

    data = normalizedImageItem.getData(copy=False)
    assert np.array_equal(ref_image, data)

    levels = normalizedImageItem.getLevels()
    assert np.allclose(levels, min_max)

    assert normalizedImageItem.quickMinMax() == min_max



# SPDX-License-Identifier: MIT

import os
import time
import traceback

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

from .widgets.UtilityWidgets import ErrorMessageBox
from . import __version__


def excepthook(exc_type, exc_value, traceback_obj):
    """
    Global function to catch unhandled exceptions. This function will result in an error dialog which displays the
    error information.

    :param exc_type: exception type
    :param exc_value: exception value
    :param traceback_obj: traceback object
    :return:
    """
    separator = '-' * 80
    log_path = f"{os.path.expanduser('~')}/dioptas_error.log"
    notice = \
        """An unhandled exception occurred. Please report the bug under:\n """ \
        """\t%s\n""" \
        """or via email to:\n\t <%s>.\n\n""" \
        """Please make sure to report the steps to reproduce the error. Otherwise it will be hard to fix it. \n\n""" \
        """A log has been written to "%s".\n\nError information:\n""" % \
        ("https://github.com/Dioptas/Dioptas/issues",
         "clemens.prescher@gmail.com", log_path)
    version_info = '\n'.join((separator, "Dioptas Version: %s" % __version__))
    time_string = time.strftime("%Y-%m-%d, %H:%M:%S")
    tb_info_file = StringIO()
    traceback.print_tb(traceback_obj, None, tb_info_file)
    tb_info_file.seek(0)
    tb_info = tb_info_file.read()
    errmsg = '%s: \n%s' % (str(exc_type), str(exc_value))
    sections = [separator, time_string, separator, errmsg, separator, tb_info]
    msg = '\n'.join(sections)
    try:
        f = open(log_path, "a")
        f.write(msg)
        f.write(version_info)
        f.close()
    except IOError:
        pass
    errorbox = ErrorMessageBox()
    errorbox.setText(str(notice) + str(msg) + str(version_info))
    errorbox.exec_()



import os
import subprocess

import numpy as np
from unittest.mock import MagicMock
import json
from dioptas.controller.MainController import MainController
from ..utility import click_button

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, os.pardir, "data")


def test_image_is_shown_correctly_after_tab_change(main_controller: MainController):
    dioptas_model = main_controller.model
    main_widget = main_controller.widget

    calibration_controller = main_controller.calibration_controller
    mask_controller = main_controller.mask_controller
    integration_controller = main_controller.integration_controller
    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))
    img_data = dioptas_model.img_model.img_data

    calibration_img_widget = calibration_controller.widget.img_widget
    assert np.array_equal(calibration_img_widget.img_data, img_data)

    click_button(main_widget.mask_mode_btn)
    mask_img_widget = mask_controller.widget.img_widget
    assert np.array_equal(mask_img_widget.img_data, img_data)

    click_button(main_widget.integration_mode_btn)
    integration_img_widget = integration_controller.widget.img_widget
    assert np.array_equal(integration_img_widget.img_data, img_data)

    click_button(main_widget.map_mode_btn)
    map_img_widget = main_controller.map_controller.widget.img_plot_widget
    assert np.array_equal(map_img_widget.img_data, img_data)

    dioptas_model.img_model.load(os.path.join(data_path, "image_002.tif"))
    img_data = dioptas_model.img_model.img_data
    click_button(main_widget.calibration_mode_btn)
    assert np.array_equal(calibration_img_widget.img_data, img_data)


def test_map_image_listening_is_enabled_after_tab_change(main_controller: MainController):
    dioptas_model = main_controller.model
    main_widget = main_controller.widget
    map_img_widget = main_controller.map_controller.widget.img_plot_widget
    click_button(main_widget.map_mode_btn)

    dioptas_model.img_model.load(os.path.join(data_path, "image_002.tif"))
    img_data = dioptas_model.img_model.img_data
    assert np.array_equal(map_img_widget.img_data, img_data)


def test_calibration_image_listener_is_enabled_after_tab_change(main_controller: MainController):
    dioptas_model = main_controller.model
    main_widget = main_controller.widget
    calibration_controller = main_controller.calibration_controller
    calibration_img_widget = calibration_controller.widget.img_widget

    click_button(main_widget.calibration_mode_btn)

    dioptas_model.img_model.load(os.path.join(data_path, "image_001.tif"))
    img_data = dioptas_model.img_model.img_data
    assert np.array_equal(calibration_img_widget.img_data, img_data)


def test_integration_image_listener_is_enabled_after_tab_change(main_controller: MainController):
    dioptas_model = main_controller.model
    main_widget = main_controller.widget
    integration_controller = main_controller.integration_controller
    integration_img_widget = integration_controller.widget.img_widget

    click_button(main_widget.integration_mode_btn)
    dioptas_model.img_model.load(os.path.join(data_path, "image_002.tif"))
    img_data = dioptas_model.img_model.img_data
    assert np.array_equal(integration_img_widget.img_data, img_data)


def test_load_with_configuration(qapp, tmp_path):
    config_file = tmp_path / "config.ini"
    config = {
        "external_actions":
            [
                {
                    "name": "A",
                    "command": "echo",
                    "arguments": "",
                },
                {
                    "name": "B",
                    "command": "sub",
                    "arguments": "-t tree",
                },
            ]
    }
    json.dump(config, open(config_file, "w"))

    main_controller = MainController(use_settings=False, config_file=config_file)
    assert main_controller.configuration == config
    assert main_controller.widget.external_action_btns["A"].text() == "A"
    assert main_controller.widget.external_action_btns["B"].text() == "B"

    img_model = main_controller.model.img_model
    img_model.load(os.path.join(data_path, "image_001.tif"))

    subprocess.Popen = MagicMock()
    click_button(main_controller.widget.external_action_btns["A"])
    subprocess.Popen.assert_called_once_with(
        [
            "echo",
            img_model.filename,
            str(img_model.series_pos),
        ],
        shell=True)

    click_button(main_controller.widget.external_action_btns["B"])
    subprocess.Popen.assert_called_with(
        [
            "sub",
            "-t",
            "tree",
            img_model.filename,
            str(img_model.series_pos),
        ],
        shell=True)


def test_title_is_shown_correctly(main_controller: MainController):
    main_controller.model.img_model.load(os.path.join(data_path, "image_001.tif"))
    title = main_controller.widget.windowTitle()
    assert title.startswith("image_001.tif | Dioptas")
    assert title.endswith("C. Prescher")

    main_controller.model.calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    main_controller.model.img_changed.emit()
    title = main_controller.widget.windowTitle()
    assert title.startswith("image_001.tif, calibration: CeO2_Pilatus1M | Dioptas")
    assert title.endswith("C. Prescher")




# SPDX-License-Identifier: MIT
from __future__ import annotations

import math
import pathlib
from typing import Optional

from qtpy import QtGui, QtCore, QtWidgets
import pyqtgraph.graphicsItems.GradientEditorItem
import numpy as np

from . import utils
from .NormalizedImageItem import NormalizedImageItem
from ... import style_path


class ColormapPopup(QtWidgets.QFrame):
    """Dialog providing control over the currently used colormap"""

    sigCurrentGradientChanged = QtCore.Signal(dict)
    """Signal emitted when the colormap gradient has changed"""

    sigCurrentNormalizationChanged = QtCore.Signal(str)
    """Signal emitted when the colormap normalization has changed"""

    sigRangeChanged = QtCore.Signal(float, float)
    """Signal emitted when the data range has changed"""

    _RESET_MODES = {  # Button text: (mode, tooltip)
        "Percentile": (
            "percentile",
            "Clip extreme values at the given percentile",
        ),
        "Min/max": ("minmax", "Use data min/max to scale colormap range"),
        "Mean±3 Std": (
            "mean3std",
            "Use data mean ± 3 × standard deviation to scale colormap range",
        ),
    }

    # Logarithmic slider parameters
    _PERCENTILE_MIN = 0.01
    _PERCENTILE_MAX = 25.0
    _PERCENTILE_SLIDER_STEPS = 200

    def __init__(self, parent=None):
        super().__init__(parent)
        self._data = None
        self.setWindowTitle("Colormap configuration")
        self.setWindowFlags(QtCore.Qt.Popup)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Raised)
        self.setWindowOpacity(0.95)
        self.setLineWidth(2)

        frameLayout = QtWidgets.QVBoxLayout(self)
        frameLayout.setContentsMargins(3, 3, 3, 3)
        frameLayout.setSpacing(3)

        colormapGroupBox = QtWidgets.QGroupBox("Colormap", self)
        frameLayout.addWidget(colormapGroupBox)
        colormapLayout = QtWidgets.QFormLayout(colormapGroupBox)
        colormapLayout.setContentsMargins(0, 6, 0, 0)
        colormapLayout.setLabelAlignment(QtCore.Qt.AlignRight)

        self._gradientComboBox = QtWidgets.QComboBox(self)
        for (
            name,
            gradient,
        ) in pyqtgraph.graphicsItems.GradientEditorItem.Gradients.items():
            icon = self._createQIconFromGradient(gradient)
            self._gradientComboBox.addItem(icon, name.capitalize(), gradient)
        self._gradientComboBox.currentIndexChanged.connect(
            self._gradientComboBoxCurrentIndexChanged
        )
        colormapLayout.addRow("Colormap:", self._gradientComboBox)

        self._normalizationComboBox = QtWidgets.QComboBox(self)
        for normalization in NormalizedImageItem.supportedNormalizations():
            description = NormalizedImageItem.getNormalizationDescription(
                normalization
            ).capitalize()
            self._normalizationComboBox.addItem(description, normalization)

        self._normalizationComboBox.setCurrentIndex(0)
        self._normalizationComboBox.currentIndexChanged.connect(
            self._normalizationComboBoxCurrentIndexChanged
        )
        colormapLayout.addRow("Normalization:", self._normalizationComboBox)

        rangeGroupBox = QtWidgets.QGroupBox("Range", self)
        frameLayout.addWidget(rangeGroupBox)
        rangeLayout = QtWidgets.QFormLayout(rangeGroupBox)
        rangeLayout.setContentsMargins(0, 6, 0, 0)
        rangeLayout.setLabelAlignment(QtCore.Qt.AlignRight)

        self._minEdit = QtWidgets.QLineEdit(self)
        self._minEdit.setValidator(QtGui.QDoubleValidator(1, float("inf"), -1))
        self._minEdit.setText("1")
        self._minEdit.editingFinished.connect(self._rangeChanged)
        rangeLayout.addRow("Min:", self._minEdit)

        self._maxEdit = QtWidgets.QLineEdit(self)
        self._maxEdit.setValidator(QtGui.QDoubleValidator(1, float("inf"), -1))
        self._maxEdit.setText("1")
        self._maxEdit.editingFinished.connect(self._rangeChanged)
        rangeLayout.addRow("Max:", self._maxEdit)

        reloadIcon = (
            QtWidgets.QApplication.instance()
            .style()
            .standardIcon(QtWidgets.QStyle.SP_BrowserReload)
        )
        self._autoscaleButton = QtWidgets.QPushButton(reloadIcon, "Reset", self)
        self._autoscaleButton.setToolTip("Scale colormap range with current mode")
        self._autoscaleButton.clicked.connect(self._autoscaleRequested)
        self._autoscaleButton.setAutoDefault(False)
        self._autoscaleButton.setEnabled(False)
        rangeLayout.addRow("", self._autoscaleButton)

        resetModeGroupBox = QtWidgets.QGroupBox("Reset Mode", self)
        frameLayout.addWidget(resetModeGroupBox)
        resetModesLayout = QtWidgets.QVBoxLayout(resetModeGroupBox)
        resetModesLayout.setContentsMargins(0, 6, 0, 0)
        resetModesLayout.setSpacing(0)

        current_mode, current_percentile = self._parse_auto_level_mode(
            utils.auto_level.mode
        )

        self._resetButtonGroup = QtWidgets.QButtonGroup(self)
        for text, (mode, tooltip) in self._RESET_MODES.items():
            radioButton = QtWidgets.QRadioButton(text, self)
            radioButton.setToolTip(tooltip)
            radioButton.setChecked(mode == current_mode)
            self._resetButtonGroup.addButton(radioButton)
            resetModesLayout.addWidget(radioButton)

            if mode == "percentile":
                sliderLayout = QtWidgets.QHBoxLayout()
                sliderLayout.setContentsMargins(20, 0, 0, 0)

                self._percentileSlider = QtWidgets.QSlider(
                    QtCore.Qt.Horizontal, self
                )
                self._percentileSlider.setRange(0, self._PERCENTILE_SLIDER_STEPS)
                self._percentileSlider.setValue(
                    self._percentile_to_slider(current_percentile)
                )
                self._percentileSlider.setToolTip("Adjust clipping percentile")
                self._percentileSlider.valueChanged.connect(
                    self._percentileSliderChanged
                )

                self._percentileLabel = QtWidgets.QLabel(self)
                self._percentileLabel.setFixedWidth(40)
                self._updatePercentileLabel()

                sliderLayout.addWidget(self._percentileSlider)
                sliderLayout.addWidget(self._percentileLabel)
                resetModesLayout.addLayout(sliderLayout)

        self._resetButtonGroup.buttonClicked.connect(self._resetModeChanged)

        filterLayout = QtWidgets.QHBoxLayout()
        filterLayout.setContentsMargins(0, 4, 0, 0)
        self._filterGapsCheckBox = QtWidgets.QCheckBox("Filter gaps", self)
        self._filterGapsCheckBox.setToolTip("Toggle detector gaps value filtering")
        self._filterGapsCheckBox.setChecked(utils.auto_level.filter_dummy)
        self._filterGapsCheckBox.toggled.connect(self._autoscaleRequested)
        filterLayout.addWidget(self._filterGapsCheckBox)
        resetModesLayout.addLayout(filterLayout)

        self._updateSliderVisibility()

        buttonBox = QtWidgets.QDialogButtonBox(parent=self)
        buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        closeButton = buttonBox.button(QtWidgets.QDialogButtonBox.Close)
        closeButton.clicked.connect(self.close)
        closeButton.setAutoDefault(False)
        frameLayout.addWidget(buttonBox)

    def setData(self, data: Optional[np.ndarray] = None, copy: bool = True):
        """Set data and histogram to use for autoscale"""
        self._data = None if data is None else np.array(data, copy=copy)
        self._autoscaleButton.setEnabled(data is not None)

    def getData(self, copy: bool = True) -> Optional[np.ndarray]:
        """Returns data used for autoscale if set else None"""
        if self._data is None:
            return None
        return np.array(self._data, copy=copy)

    def _gradientComboBoxCurrentIndexChanged(self, index: int):
        if index < 0:
            return
        gradient = self._gradientComboBox.itemData(index, QtCore.Qt.UserRole)
        self.sigCurrentGradientChanged.emit(gradient)

    def _normalizationComboBoxCurrentIndexChanged(self, index: int):
        if index < 0:
            return
        normalization = self._normalizationComboBox.itemData(index, QtCore.Qt.UserRole)
        self.sigCurrentNormalizationChanged.emit(normalization)

    def setCurrentGradient(self, gradient: dict):
        """Set the currently selected gradient

        If the gradient is not available, a 'Custom' item is added for it.
        """
        for (
            name,
            description,
        ) in pyqtgraph.graphicsItems.GradientEditorItem.Gradients.items():
            if (
                gradient["mode"] == description["mode"]
                and gradient["ticks"] == description["ticks"]
            ):
                self._gradientComboBox.setCurrentText(name.capitalize())
                return

        icon = self._createQIconFromGradient(gradient)
        # Block signals to avoid emitting with previously selected gradient since index changes
        wasBlocked = self._gradientComboBox.blockSignals(True)
        self._gradientComboBox.insertItem(0, icon, "Custom", gradient)
        self._gradientComboBox.blockSignals(wasBlocked)
        self._gradientComboBox.setCurrentIndex(0)

    def getCurrentGradient(self) -> dict:
        """Returns the currently selected gradient"""
        return self._gradientComboBox.currentData()

    def setCurrentNormalization(self, normalization: str):
        """Set the currently selected normalization"""
        index = self._normalizationComboBox.findData(normalization)
        if index < 0:
            raise ValueError(f"Unsupported normalization: {normalization}")
        self._normalizationComboBox.setCurrentIndex(index)

    def getCurrentNormalization(self) -> str:
        """Returns the currently selected normalization"""
        return self._normalizationComboBox.currentData()

    def _rangeChanged(self):
        minimum, maximum = self.getRange()
        if maximum < minimum:
            self.setRange(maximum, minimum)
            return
        self.sigRangeChanged.emit(minimum, maximum)

    def setRange(self, minimum: float, maximum: float):
        """Set the data range (min, max) of the colormap"""
        if maximum < minimum:
            minimum, maximum = maximum, minimum
        if np.allclose((minimum, maximum), self.getRange()):
            return
        self._minEdit.setText(
            self._minEdit.validator().locale().toString(float(minimum))
        )
        self._maxEdit.setText(
            self._maxEdit.validator().locale().toString(float(maximum))
        )
        self._rangeChanged()

    def getRange(self) -> tuple[float, float]:
        """Returns the data range of the colormap (min, max)"""
        minimum, validated = (
            self._minEdit.validator().locale().toDouble(self._minEdit.text())
        )
        if not validated:
            minimum = 1
        maximum, validated = (
            self._maxEdit.validator().locale().toDouble(self._maxEdit.text())
        )
        if not validated:
            maximum = minimum
        return minimum, maximum

    def _getResetMode(self) -> str:
        button = self._resetButtonGroup.checkedButton()
        if button is not None and button.text() in self._RESET_MODES:
            mode = self._RESET_MODES[button.text()][0]
            if mode == "percentile":
                percentile = self._slider_to_percentile(
                    self._percentileSlider.value()
                )
                return f"{percentile}percentile"
            return mode
        return "percentile"  # Fallback

    @staticmethod
    def _parse_auto_level_mode(mode: str) -> tuple[str, float]:
        """Parse an AutoLevel mode string into (reset_mode, percentile).

        Returns e.g. ("percentile", 1.5) for "1.5percentile",
        ("minmax", 0.4) for "minmax", ("percentile", 0.4) for "default".
        """
        import re

        match = re.match(r"(\d+(?:\.\d*)?)percentile", mode)
        if match is not None:
            return "percentile", float(match.group(1))
        if mode in ("minmax", "mean3std"):
            return mode, 0.4
        # "default" or unknown → percentile with default value
        return "percentile", 0.4

    def _percentile_to_slider(self, percentile: float) -> int:
        """Convert a percentile value to slider position (logarithmic)."""
        percentile = max(self._PERCENTILE_MIN, min(self._PERCENTILE_MAX, percentile))
        pos = self._PERCENTILE_SLIDER_STEPS * math.log(
            percentile / self._PERCENTILE_MIN
        ) / math.log(self._PERCENTILE_MAX / self._PERCENTILE_MIN)
        return round(pos)

    def _slider_to_percentile(self, pos: int) -> float:
        """Convert a slider position to percentile value (logarithmic)."""
        value = self._PERCENTILE_MIN * (
            self._PERCENTILE_MAX / self._PERCENTILE_MIN
        ) ** (pos / self._PERCENTILE_SLIDER_STEPS)
        return round(value, 2)

    def _resetModeChanged(self, button):
        """Handle reset mode radio button change."""
        self._updateSliderVisibility()
        self._autoscaleRequested()

    def _updateSliderVisibility(self):
        """Show the percentile slider only when Percentile mode is selected."""
        button = self._resetButtonGroup.checkedButton()
        visible = button is not None and button.text() == "Percentile"
        self._percentileSlider.setVisible(visible)
        self._percentileLabel.setVisible(visible)

    def _percentileSliderChanged(self, value: int):
        """Handle slider value change: update label and re-trigger autoscale."""
        self._updatePercentileLabel()
        button = self._resetButtonGroup.checkedButton()
        if button is not None and button.text() == "Percentile":
            self._autoscaleRequested()

    def _updatePercentileLabel(self):
        """Update the label next to the slider with the current percentile."""
        percentile = self._slider_to_percentile(self._percentileSlider.value())
        self._percentileLabel.setText(f"{percentile}%")

    def _autoscaleRequested(self, *args):
        utils.auto_level.mode = self._getResetMode()
        utils.auto_level.filter_dummy = self._filterGapsCheckBox.isChecked()
        colormapRange = utils.auto_level.get_range(self.getData(copy=False))
        if colormapRange is None:
            return
        self.setRange(*colormapRange)

    @staticmethod
    def _createQIconFromGradient(gradient: dict) -> QtGui.QIcon:
        """Generates a QIcon from a pyqtgraph gradient"""
        gradientEditorItem = (
            pyqtgraph.graphicsItems.GradientEditorItem.GradientEditorItem()
        )
        gradientEditorItem.setLength(100)
        gradientEditorItem.restoreState(gradient)
        qgradient = gradientEditorItem.getGradient()

        pixmap = QtGui.QPixmap(100, 100)
        painter = QtGui.QPainter(pixmap)
        brush = QtGui.QBrush(qgradient)
        painter.fillRect(QtCore.QRect(0, 0, 100, 100), brush)
        painter.end()
        return QtGui.QIcon(pixmap)
